import os
from selector.ta_execution import tae_from_cmd_wrapper_rt, tae_from_cmd_wrapper_quality
import time
import logging

def get_tournament_membership(tournaments, conf):
    """
    For a list of tournaments, determine of which a conf is a member.
    :param tournaments: List
    :param conf: Conf
    :return:
    """
    for t in tournaments:
        if conf.id in t.configuration_ids or conf.id in t.worst_finisher or conf.id in t.best_finisher:
            return t

def get_get_tournament_membership_with_ray_id(task_id, tournaments):
    """
    For a ray task id return the tournament it belongs to
    :param task_id:
    :param tournaments:
    :return:
    """

    ob_t = None
    for t in tournaments:
        t_objects = t.ray_object_store
        for confs, instance_objects in t_objects.items():
            for inst, ob in instance_objects.items():
                if ob == task_id:
                    ob_t = t
                    pass
    return ob_t

def get_tasks(taskdic, tasks):
    """
    Map back a ray object to the conf/instance pair.
    :param taskdic: Nested dic of {conf: {instance: ray object}}
    :param tasks: List with ray objects that are running
    :return: List of [conf, instances] pairs that are currently running
    """
    running_tasks = []
    for conf, instance in taskdic.items():
        for instance_name, object in instance.items():
            if object in tasks:
                running_tasks.append([conf, instance_name])
    return running_tasks

def update_tasks(tasks, next_task, tournament, global_cache, ta_wrapper, scenario):
    """

    :param tasks: List of ray objects
    :param next_task: List of [conf, instance] pairs
    :param tournament: Tournament the next task is part of
    :param global_cache: Ray cache
    :param ta_wrapper:
    :param scenario:
    :return: Updated list of ray objects
    """
    for t in next_task:
        if t[1] is not None:
            # TODO need to change the wrapper to something more generic here
            if scenario.run_obj == "runtime":
                task = tae_from_cmd_wrapper_rt.remote(t[0], t[1], global_cache, ta_wrapper, scenario)
            elif scenario.run_obj == "quality":
                task = tae_from_cmd_wrapper_quality.remote(t[0], t[1], global_cache, ta_wrapper, scenario)
            tasks.append(task)
            # We also add the ray object id to the tournament to latter map the id back
            if t[0].id not in tournament.ray_object_store.keys():
                tournament.ray_object_store[t[0].id] = {t[1]: task}
            else:
                tournament.ray_object_store[t[0].id][t[1]] = task
    return tasks


def termination_check(termination_criterion, main_loop_start, total_runtime, total_tournament_number,
                      tournament_counter):
    """
    Check what termination criterion for the main tournament loop has been parsed and return true,
    if the criterion is not met yet.
    :param termination_criterion: Str. termination criterion for the tournament main loop
    :param main_loop_start: Int. Time of the start of the tournament main loop
    :param total_runtime: Int. Total runtime for the main loop, when the termination criterion is "total_runtime"
    :param total_tournament_number: Int. Total number of tournaments for the main loop,
                                    when the termination criterion is "total_tournament_number"
    :param tournament_counter: Int. Number of tournaments, that finished already
    :return: Bool. True, when the termination criterion is not met, False otherwise
    """
    if termination_criterion == "total_runtime":
        return time.time() - main_loop_start < total_runtime

    elif termination_criterion == "total_tournament_number":
        return tournament_counter < total_tournament_number

    else:
        return time.time() - main_loop_start < total_runtime






"""Main module of selector."""

import sys
import os
sys.path.append(os.getcwd())

os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"

import importlib
import logging
import numpy as np
import ray

from selector.scenario import Scenario, parse_args
from selector.log_setup import clear_logs, check_log_folder, save_latest_logs
from selector.mini_tournaments import offline_mini_tournament_configuration
from selector.best_conf import safe_best

sys.path.append(os.getcwd())


if __name__ == "__main__":
    selector_args = parse_args()

    wrapper_mod = importlib.import_module(selector_args["wrapper_mod_name"])

    wrapper_name = selector_args["wrapper_class_name"]
    wrapper_ = getattr(wrapper_mod, wrapper_name)
    ta_wrapper = wrapper_()

    scenario = Scenario(selector_args["scenario_file"], selector_args)

    np.random.seed(scenario.seed)

    check_log_folder(scenario.log_folder)
    clear_logs(scenario.log_folder)

    logging.\
        basicConfig(level=logging.INFO,
                    format='%(asctime)s %(message)s',
                    handlers=[logging.FileHandler(
                        f"./selector/logs/{scenario.log_folder}/main.log"), ])

    logger = logging.getLogger(__name__)

    logger.info(f"Logging to {scenario.log_folder}")
    # print(scenario.parameter)

    # init: ray.init(address="auto") for cluster, ray.init() for PC
    # ray.init(address="auto")
    ray.init()

    logger.info("Ray info: {}".format(ray.cluster_resources()))
    logger.info("Ray nodes {}".format(ray.nodes()))
    logger.info("WD: {}".format(os.getcwd()))

    offline_mini_tournament_configuration(scenario, ta_wrapper, logger)

    save_latest_logs(scenario.log_folder)
    if scenario.termination_criterion == 'runtime':
        safe_best(f'./selector/logs/{scenario.log_folder}/',
                  scenario.cutoff_time)
    elif scenario.termination_criterion == 'quality':
        safe_best(f'./selector/logs/{scenario.log_folder}/',
                  sys.maxsize)
    ray.shutdown()



import numpy as np
import ray
import uuid
import time
import copy

from selector.pool import Tournament
from selector.tournament_performance import get_censored_runtime_for_instance_set, get_instances_no_results, get_conf_time_out, get_runtime_for_instance_set_with_timeout

class MiniTournamentDispatcher:

    def init_tournament(self, results, configurations, instance_partition, instance_partition_id):
        """
        Create a new tournament out of the given configurations and list of instances.
        :param results: Results cache.
        :param configurations: List. Configurations for the tournament
        :param instance_partition: List. List of instances
        :param instance_partition_id: Id of the instance set.
        :return: Tournament, first conf/instance assignment to run
        """
        # Get the configuration that has seen the most instances before
        conf_instances_ran = []
        most_run_conf = None
        for conf in configurations:
            if conf.id in list(results.keys()):
                conf_instances_ran = list(results[conf.id].keys())
                most_run_conf = conf

        # Get instances the conf with the most runs has not been run on before
        possible_first_instances = [i for i in instance_partition if i not in conf_instances_ran]

        # If there are instances the conf with the most runs has not seen we select on of them to be the first instance
        # all confs should be run on
        if possible_first_instances:
            first_instance = np.random.choice(possible_first_instances)
            initial_instance_conf_assignments = [[conf, first_instance] for conf in configurations]
            best_finisher = []
        # An empty list of possible instances means that the conf with the most runs has seen all instances in the
        # instance set. In that case we can choose any instance for the confs that have not seen all instances.
        # We also have a free core then to which we assign a extra conf/instance pair where both are chosen at random
        else:
            first_instance = np.random.choice(instance_partition)

            #configurations_not_run_on_all = copy.deepcopy(configurations)
            configurations_not_run_on_all = configurations
            configurations_not_run_on_all.remove(most_run_conf)

            extra_instances = copy.deepcopy(instance_partition)
            extra_instances.remove(first_instance)

            extra_assignment = [np.random.choice(configurations_not_run_on_all), np.random.choice(extra_instances)]
            initial_instance_conf_assignments = [[conf, first_instance] for conf in configurations_not_run_on_all] \
                                                + [extra_assignment]

            best_finisher = [most_run_conf]

        configuration_ids = [c.id for c in configurations] + [b.id for b in best_finisher if len(best_finisher) >= 1]

        return Tournament(uuid.uuid4(), best_finisher, [], configurations, configuration_ids, {}, instance_partition,
                          instance_partition_id), \
               initial_instance_conf_assignments


    def update_tournament(self, results, tasks, finished_conf, tournament, number_winner, time_out, par_penalty):
        """
        Given a finishing conf we update the tournament if necessary. I.e the finishing conf has seen all instances of
        the tournament. In that case, it is moved either to the best or worst finishers. best finishers are ordered.
        Worst finishers are not
        :param results: Ray cache object.
        :param finished_conf: Configuration that finished or was canceled
        :param tournament: Tournament the finish conf was a member of
        :param number_winner: Int that determines the number of winners per tournament
        :return: updated tournament, stopping signal
        """
        evaluated_instances = results[finished_conf.id].keys()
        bfi_add = False

        # We figure out if there are still tasks the finished configuration is still running on for which we have a results
        # but have not returned through a ray.wait()
        still_running_task_for_conf = [sr for sr in tasks if sr in list(tournament.ray_object_store[finished_conf.id].values())]

        # A conf can only become a best finisher if it has seen all instances of the tournament and is not running any
        # other conf/instance pairs. i.e the result we process here is the last one
        if set(evaluated_instances) == set(tournament.instance_set) and len(still_running_task_for_conf) == 0:
            # We can than remove the conf from further consideration
            if finished_conf in tournament.configurations:
                tournament.configurations.remove(finished_conf)
            else:
                print(f"conf not {finished_conf}")
            finished_conf_runtime_mean = get_runtime_for_instance_set_with_timeout(results, finished_conf.id,
                                                                                   tournament.instance_set, time_out, par_penalty) / len(tournament.instance_set)

            # If there are already some best finisher we need to compare the conf to them
            if len(tournament.best_finisher) > 0:
                # We assume that the finishers in the set are ordered according to their runtime
                for bfi in range(len(tournament.best_finisher)):
                    bfrm = get_runtime_for_instance_set_with_timeout(results, tournament.best_finisher[bfi].id,
                                                                     tournament.instance_set, time_out, par_penalty) / len(tournament.instance_set)

                    # If the conf is better than one best finisher we insert it
                    if finished_conf_runtime_mean <= bfrm:
                        tournament.best_finisher.insert(bfi, finished_conf)
                        bfi_add = True
                        if finished_conf in tournament.worst_finisher:
                            tournament.worst_finisher.remove(finished_conf)
                        # If we have too many best finishers we cut off the excess
                        if len(tournament.best_finisher) > number_winner:
                            transition =  number_winner - len(tournament.best_finisher)
                            tournament.worst_finisher = tournament.worst_finisher + tournament.best_finisher[transition:]
                            tournament.best_finisher = tournament.best_finisher[: transition]
                            break
                # We also add a conf to best finishers if we have not enough
                if len(tournament.best_finisher) < number_winner and not bfi_add:
                    tournament.best_finisher.append(finished_conf)

                # If the conf is not better it is a worst finisher
                elif finished_conf not in tournament.worst_finisher and not bfi_add:
                    tournament.worst_finisher.append(finished_conf)
            else:
                tournament.best_finisher.append(finished_conf)


        # If there are no configurations left we end the tournament
        if len(tournament.configurations) == 0:
            stop = True
        else:
            stop = False

        return tournament, stop

    def next_tournament_run(self, results, tournament, finished_conf):
        """
        Decided which conf/instance pair to run next. Rule: If the configuration that has just finished was not killed
        nor saw all instances, it is assigned a new instance at random. Else, the configuration with the lowest runtime
        so far is selected.
        :param results: Ray cache
        :param tournament: The tournament we opt to create a new task for
        :param finished_conf: Configuration that just finished before
        :return: configuration, instance pair to run next
        """
        next_possible_conf = {}

        # For each conf still in the running we need to figure out on which instances it already ran or is still
        # running on to get for each conf the instances it still can run on
        for conf in tournament.configurations:
            already_run = get_instances_no_results(results, conf.id, tournament.instance_set)

            not_running_currently = get_instances_no_results(tournament.ray_object_store, conf.id,
                                                             tournament.instance_set)
            not_running_currently = [c for c in not_running_currently if c in already_run]

            if len(not_running_currently) > 0:
                next_possible_conf[conf.id] = not_running_currently
        # If there are no configuration that need to see new instances we create a dummy task to give the still running
        # conf/instance pairs time to finish.
        if len(next_possible_conf) == 0:
            configuration = None
            next_instance = None
        else:
        # If the previous run conf has not seen all instances and did not time out it is selected to run again
            if finished_conf.id in list(next_possible_conf.keys()):
                next_conf_id = finished_conf.id
            else: #Select the configuration with the lowest mean runtime
                mean_rt_store = {}
                for conf in next_possible_conf.keys():
                    if conf in results.keys():
                        conf_rt = list(results[conf].values())
                        mean_rt_store[conf] = sum(conf_rt) / len(conf_rt)
                if mean_rt_store:
                    next_conf_id = min(mean_rt_store, key=mean_rt_store.get)
                # In case we have no results for any of the remaining configuration we sample
                else:
                    next_conf_id = np.random.choice(list(next_possible_conf.keys()))

            configuration = [c for c in tournament.configurations if c.id == next_conf_id][0]
            next_possible_instance = next_possible_conf[next_conf_id]
            next_instance = np.random.choice(next_possible_instance)

        return [[configuration, next_instance]]


"""This modules includes functions to execute the target algorithm."""


import logging
import time
import subprocess
import ray
import numpy as np
import psutil
import os
import signal
import traceback
import re
import copy

from threading import Thread
from queue import Queue, Empty
from selector.generators.default_point_generator import check_conditionals


def enqueue_output(out, queue):
    """Enqueue output."""
    for line in iter(out.readline, b''):
        line = line.decode("utf-8")
        queue.put(line)
    out.close()


def get_running_processes(ta_process_name):
    """Get list of running processes."""
    processes = []
    for proc in psutil.process_iter():
        try:
            processName = proc.name()
            processID = proc.pid
            if processName in [ta_process_name]:
                processes.append([processName, processID])
        except (psutil.NoSuchProcess, psutil.AccessDenied,
                psutil.ZombieProcess):
            pass
    return processes


def termination_check(process_pid, process_status, ta_process_name, python_pid,
                      conf_id, instance):
    """Check if process was terminated."""
    running_processes = get_running_processes(ta_process_name)

    sr = False
    for rp in running_processes:
        if process_pid == rp[1]:
            sr = True

    if sr:
        logging.info(f"Failed to terminate {conf_id}, {instance}: process {process_pid} with {process_status} on {python_pid} is still running")
    else:
        logging.info(
            f"Successfully terminated {conf_id}, {instance} on {python_pid} with {process_status}")


@ray.remote(num_cpus=1)  # , max_retries=0,  retry_exceptions= False)
def tae_from_cmd_wrapper_rt(conf, instance_path, cache, ta_command_creator, scenario):
    """
    Execute the target algorithm with a given conf/instance pair by calling a user provided Wrapper that created a cmd
    line argument that can be executed
    :param conf: Configuration
    :param instance: Instances
    :param cache: Cache
    :param ta_command_creator: Wrapper that creates a
    :return:
    """
    # todo logging dic should be provided somewhere else -> DOTAC-37
    logging.basicConfig(
        filename=f'./selector/logs/{scenario.log_folder}/wrapper_log_for{conf.id}.log',
        level=logging.INFO, format='%(asctime)s %(message)s')

    try:
        logging.info(f"Wrapper TAE start {conf}, {instance_path}")
        runargs = {'instance': f'{scenario.instances_dir + instance_path}',
                   'seed': scenario.seed if scenario.seed else -1,
                   "id": f"{conf.id}"}

        clean_conf = copy.copy(conf.conf)
        # Check conditionals and turn off parameters if violated
        cond_vio = check_conditionals(scenario, clean_conf)
        for cv in cond_vio:
            clean_conf.pop(cv, None)
        cmd = ta_command_creator.get_command_line_args(runargs, clean_conf)
        start = time.time()
        cache.put_start.remote(conf.id, instance_path, start)

        p = psutil.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT, close_fds=True)

        q = Queue()
        t = Thread(target=enqueue_output, args=(p.stdout, q))
        t.daemon = True
        t.start()

        timeout = False
        empty_line = False
        memory_p = 0
        cpu_time_p = 0
        reading = True
        solved = False

        while reading:
            try:
                line = q.get(timeout=.5)
                empty_line = False
                # Get the cpu time and memory of the process
                while cpu_time_p == 0.0:
                    cpu_times = p.cpu_times()
                    if "children_user" in cpu_times._fields:
                        cpu_time_p = cpu_times.children_user
                    else:
                        cpu_time_p = cpu_times.user
                memory_p = p.memory_info().rss / 1024 ** 2
            except Empty:
                empty_line = True
                if p.poll() is None:
                    # cpu_time_p = time.time() - start
                    cpu_times = p.cpu_times()
                    if "children_user" in cpu_times._fields:
                        cpu_time_p = cpu_times.children_user
                    else:
                        cpu_time_p = cpu_times.user
                    memory_p = p.memory_info().rss / 1024 ** 2
                if float(cpu_time_p) > float(scenario.cutoff_time) or float(memory_p) > float(scenario.memory_limit) and timeout ==False:
                    timeout = True
                    logging.info(f"Timeout or memory reached, terminating: {conf}, {instance_path} {time.time() - start}")
                    print(f"Timeout or memory reached, terminating: {conf}, {instance_path} {time.time() - start}")
                    if p.poll() is None:
                        p.terminate()
                    try:
                        time.sleep(1)
                    except:
                        print("Got sleep interupt", conf, instance_path)
                        pass
                    if p.poll() is None:
                        p.kill()
                    try:
                        os.killpg(p.pid, signal.SIGKILL)
                    except Exception:
                        pass
                    # if scenario.ta_pid_name is not None:
                    #    termination_check(p.pid, p.poll(), scenario.ta_pid_name, os.getpid(),conf.id, instance_path)
                pass
            else:  # write intemediate feedback
                # print(line)
                if "placeholder" in line:
                    cache.put_intermediate_output.remote(conf.id, instance_path, line)
                    logging.info(f"Wrapper TAE intermediate feedback {conf}, {instance_path} {line}")

                if scenario.output_trigger:
                    if scenario.solve_match in line:
                        # print('\nSolved!\n', line, '\n')
                        solved = True

            if p.poll() is None:
                # Get the cpu time and memory of the process
                cpu_times = p.cpu_times()
                if "children_user" in cpu_times._fields:
                    cpu_time_p = cpu_times.children_user
                else:
                    cpu_time_p = cpu_times.user
                memory_p = p.memory_info().rss / 1024 ** 2

                if float(cpu_time_p) > float(scenario.cutoff_time) or float(memory_p) > float(
                        scenario.memory_limit) and timeout is False:
                    timeout = True
                    logging.info(f"Timeout or memory reached, terminating: {conf}, {instance_path} {time.time() - start}")
                    if p.poll() is None:
                        p.terminate()
                    try:
                        time.sleep(1)
                    except:
                        print("Got sleep interupt", conf, instance_path)
                        pass
                    if p.poll() is None:
                        p.kill()
                    try:
                        os.killpg(p.pid, signal.SIGKILL)
                    except Exception:
                        pass

            # Break the while loop when the ta was killed or finished
            if empty_line and p.poll() is not None:
                reading = False

        if timeout:
            cache.put_result.remote(conf.id, instance_path, np.nan)
        elif scenario.output_trigger:
            if solved:
                cache.put_result.remote(conf.id, instance_path, cpu_time_p)
            else:
                cache.put_result.remote(conf.id, instance_path, np.nan)
        else:
            cache.put_result.remote(conf.id, instance_path, cpu_time_p)

        time.sleep(0.2)
        logging.info(f"Wrapper TAE end {conf}, {instance_path}")
        return conf, instance_path, False

    except KeyboardInterrupt:
        logging.info(f" Killing: {conf}, {instance_path} ")
        # We only terminated the subprocess in case it has started (p is defined)
        if 'p' in vars():
            if p.poll() is None:
                p.terminate()
            try:
                time.sleep(1)
            except:
                print("Got sleep interupt", conf, instance_path)
                pass
            if p.poll() is None:
                p.kill()
            try:
                os.killpg(p.pid, signal.SIGKILL)
            except Exception as e:
                pass
            # if scenario.ta_pid_name is not None:
            #   termination_check(p.pid, p.poll(), scenario.ta_pid_name, os.getpid(), conf.id, instance_path)
        cache.put_result.remote(conf.id, instance_path, np.nan)
        try:
            logging.info(f"Killing status: {p.poll()} {conf.id} {instance_path}")
        except:
            pass
        return conf, instance_path, True
    except Exception:
        print({traceback.format_exc()})
        logging.info(f"Exception in TA execution: {traceback.format_exc()}")


@ray.remote(num_cpus=1)
def tae_from_cmd_wrapper_quality(conf, instance_path, cache, ta_command_creator, scenario):
    """
    Execute the target algorithm with a given conf/instance pair by calling a user provided Wrapper that created a cmd
    line argument that can be executed
    :param conf: Configuration
    :param instance: Instances
    :param cache: Cache
    :param ta_command_creator: Wrapper that creates a
    :return:
    """
    logging.basicConfig(filename=f'./selector/logs/{scenario.log_folder}/wrapper_log_for{conf.id}.log', level=logging.INFO,
                        format='%(asctime)s %(message)s')

    try:
        logging.info(f"Wrapper TAE start {conf}, {instance_path}")
        runargs = {'instance': f'{scenario.instances_dir + instance_path}',
                   'seed': scenario.seed if scenario.seed else -1,
                   "id": f"{conf.id}", "timeout": scenario.cutoff_time}

        clean_conf = copy.copy(conf.conf)
        # Check conditionals and turn off parameters if violated
        cond_vio = check_conditionals(scenario, clean_conf)
        for cv in cond_vio:
            clean_conf.pop(cv, None)

        cmd = ta_command_creator.get_command_line_args(runargs, conf.conf)
        start = time.time()
        cache.put_start.remote(conf.id, instance_path, start)

        p = psutil.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT, close_fds=True)

        q = Queue()
        t = Thread(target=enqueue_output, args=(p.stdout, q))
        t.daemon = True
        t.start()

        timeout = False
        empty_line = False
        memory_p = 0
        cpu_time_p = 0
        reading = True
        quality = [np.nan]

        while reading:
            try:
                line = q.get(timeout=.5)
                empty_line = False
                # Get the cpu time and memory of the process
            except Empty:
                empty_line = True
                pass
            else:  # write intemediate feedback
                if "placeholder" in line:
                    cache.put_intermediate_output.remote(conf.id, instance_path, line)
                    logging.info(f"Wrapper TAE intermediate feedback {conf}, {instance_path} {line}")

                if scenario.run_obj == "quality":
                    output_tigger = re.search(scenario.quality_match, line)
                    if output_tigger:
                        quality = re.findall(f"{scenario.quality_extract}", line)

            # Break the while loop when the ta was killed or finished
            if empty_line and p.poll() is not None:
                reading = False

        cache.put_result.remote(conf.id, instance_path, float(quality[0]))

        time.sleep(0.2)
        logging.info(f"Wrapper TAE end {conf}, {instance_path}")
        return conf, instance_path, False
    except Exception:
        logging.info(f"Exception in TA execution: {traceback.format_exc()}")


@ray.remote(num_cpus=1)
def dummy_task(conf, instance_path, cache):
    time.sleep(2)
    cache.put_result.remote(conf.id, instance_path, np.nan)
    return conf, instance_path, True


@ray.remote(num_cpus=1)
def tae_from_aclib(conf, instance, cache, ta_exc):
    pass
# TODO



import logging
import sys
import os
sys.path.append(os.getcwd())

import numpy as np
import ray
import time
from sklearn_extra.cluster import KMedoids


from selector.scenario import Scenario
from selector.pointselector import HyperparameterizedSelector
from selector.ta_result_store import TargetAlgorithmObserver
from selector.ta_execution import dummy_task

from selector.point_gen import PointGen
from selector.pool import Surrogates
from selector.generators.random_point_generator import random_point
from selector.generators.default_point_generator import default_point
from selector.generators.variable_graph_point_generator import variable_graph_point, Mode
from selector.generators.lhs_point_generator import lhc_points, LHSType, Criterion
from selector.selection_features import FeatureGenerator
from selector.generators.surrogates.surrogates import SurrogateManager
# from selector.surrogates.surrogates import SurrogateManager

from selector.tournament_dispatcher import MiniTournamentDispatcher
from selector.tournament_bookkeeping import get_tournament_membership, update_tasks, get_tasks, termination_check, get_get_tournament_membership_with_ray_id
from selector.log_setup import clear_logs, log_termination_setting, check_log_folder, save_latest_logs

from selector.tournament_monitor import Monitor
from selector.tournament_performance import overall_best_update, get_instances_no_results

# from selector.wrapper.tap_work_wrapper import TAP_Work_Wrapper
from selector.instance_sets import TimedInstanceSet
from selector.instance_monitor import InstanceMonitor
from selector.best_conf import safe_best


def offline_mini_tournament_configuration(scenario, ta_wrapper, logger):
    log_termination_setting(logger, scenario)

    hp_seletor = HyperparameterizedSelector()
    tournament_dispatcher = MiniTournamentDispatcher()
    global_cache = TargetAlgorithmObserver.remote(scenario)
    if scenario.run_obj == "runtime":
        if scenario.monitor == "tournament_level":
            monitor = Monitor.remote(1, global_cache, scenario)
        elif scenario.monitor == "instance_level":
            monitor = InstanceMonitor.remote(1, global_cache, scenario)
        monitor.monitor.remote()

    random_generator = PointGen(scenario, random_point)
    default_point_generator = PointGen(scenario, default_point)
    vg_point_generator = PointGen(scenario, variable_graph_point)
    lhc_point_generator = PointGen(scenario, lhc_points)

    instance_selector = TimedInstanceSet(scenario.instance_set, scenario.initial_instance_set_size, scenario.set_size, runtime=scenario.wallclock_limit, start_time=0.15, end_time=0.7)
    tasks = []
    tournaments = []
    tournament_counter = 0
    results = ray.get(global_cache.get_results.remote())

    # creating the first tournaments and adding first conf/instance pairs to ray tasks
    for tc in range(scenario.number_tournaments):
        if tc == 0:
            points_to_run = [random_generator.point_generator() for _ in range(scenario.tournament_size-1)] + [default_point_generator.point_generator()]
        else:
            points_to_run = [random_generator.point_generator() for _ in range(scenario.tournament_size)]

        instance_id, instances = instance_selector.get_subset(0, 0, 0)
        tournament, initial_assignments = tournament_dispatcher.init_tournament(results, points_to_run,
                                                                                instances, instance_id)
        tournaments.append(tournament)
        #global_cache.put_tournament_history.remote(tournament)
        global_cache.put_tournament_update.remote(tournament)
        tasks = update_tasks(tasks, initial_assignments, tournament, global_cache, ta_wrapper, scenario)

    #starting the monitor

    logger.info(f"Initial Tournaments {tournaments}")
    logger.info(f"Initial Tasks, {[get_tasks(o.ray_object_store, tasks) for o in tournaments]}")

    main_loop_start = time.time()
    epoch = 0
    max_epochs = 256

    inject_default = 0

    cutoff_time = scenario.cutoff_time
    predicted_quals = []
    evaluated = []
    qap = False

    fg = FeatureGenerator(logger=None)
    sm = SurrogateManager(scenario, logger=logger)

    bug_handel = []
    tournament_history = {}
    surrogate_amortized_time = 20
    next_surrogate_update = 1
    surrogate_update_counter = 1

    while termination_check(scenario.termination_criterion, main_loop_start, scenario.wallclock_limit,
                            scenario.total_tournament_number, tournament_counter):

        winner, not_ready = ray.wait(tasks)
        tasks = not_ready
        try:
            result = ray.get(winner)[0]
            result_conf, result_instance, cancel_flag = result[0], result[1], result[2]

        # Some time a ray worker may crash. We handel that here. I.e if the TA did not run to the end, we reschedule
        except (ray.exceptions.WorkerCrashedError, ray.exceptions.TaskCancelledError, ray.exceptions.RayTaskError) as e:
            logger.info(f'Crashed TA worker, {time.ctime()}, {winner}, {e}')
            # Figure out which tournament conf. belongs to
            for t in tournaments:
                conf_instance = get_tasks(t.ray_object_store, winner)
                if len(conf_instance) != 0:
                    tournament_of_c_i = t
                    break

            conf = [conf for conf in tournament_of_c_i.configurations if conf.id == conf_instance[0][0]][0]
            instance = conf_instance[0][1]
            # We check if we have killed the conf and only messed up the termination of the process

            termination_check_c_i = ray.get(global_cache.get_termination_single.remote(conf.id , instance))
            if termination_check_c_i:
                result_conf = conf
                result_instance = instance
                cancel_flag = True
                global_cache.put_result.remote(result_conf.id, result_instance, np.nan)
                logger.info(f"Canceled task with no return: {result_conf}, {result_instance}")
                print(f"Canceled task with no return: {result_conf}, {result_instance}")
            else:  # got no results: need to rescheulde
                next_task = [[conf, instance]]
                tasks = update_tasks(tasks, next_task, tournament_of_c_i, global_cache, ta_wrapper, scenario)
                logger.info(f"We have no results: rescheduling {conf.id}, {instance} {[get_tasks(o.ray_object_store, tasks) for o in tournaments]}")
                print(f"We have no results: rescheduling {conf.id}, {instance} {[get_tasks(o.ray_object_store, tasks) for o in tournaments]}")
                continue

        # Getting the tournament of the first task id
        if len(tasks) > 0:
            first_task = tasks[0]
            ob_t = get_get_tournament_membership_with_ray_id(first_task, tournaments)

            # Figure out if the tournament of the first task is stale. If so cancel the task and start dummy task.
            if ob_t is not None:
                if len(ob_t.configurations) == 1:
                    i_no_result = get_instances_no_results(results, ob_t.configurations[0].id, ob_t.instance_set)
                    if len(i_no_result) == 1:
                        termination = ray.get(global_cache.get_termination_single.remote(ob_t.configurations[0].id, i_no_result[0]))
                        result = ray.get(global_cache.get_results_single.remote(ob_t.configurations[0].id, i_no_result[0]))
                        if termination and result == False and [ob_t.configurations[0],i_no_result[0]] not in bug_handel:
                            logger.info(f"Stale tournament: {time.strftime('%X %x %Z')}, {ob_t.configurations[0]}, {i_no_result[0]} , {first_task}, {bug_handel}")
                            print(f"Stale tournament: {time.strftime('%X %x %Z')}, {ob_t.configurations[0]}, {i_no_result[0]} , {first_task}, {bug_handel}")
                            ready_ids, _remaining_ids = ray.wait([first_task], timeout=0)
                            if len(_remaining_ids) == 1:
                                ray.cancel(first_task)
                                tasks.remove(first_task)
                                task = dummy_task.remote(ob_t.configurations[0],i_no_result[0], global_cache)
                                tasks.append(task)
                                bug_handel.append([ob_t.configurations[0],i_no_result[0]])

        if result_conf.id in list(results.keys()):
            results[result_conf.id][result_instance] = ray.get(global_cache.get_results_single.remote(result_conf.id,result_instance ))
        else:
            results[result_conf.id]= {}
            results[result_conf.id][result_instance] = ray.get(global_cache.get_results_single.remote(result_conf.id,result_instance ))

        result_tournament = get_tournament_membership(tournaments, result_conf)

        # Check whether we canceled a task or if the TA terminated regularly
        # In case we canceled a task, we need to remove it from the ray tasks
        if cancel_flag:
            if result_conf.id in result_tournament.ray_object_store.keys():
                if result_instance in result_tournament.ray_object_store[result_conf.id ].keys():
                    if result_tournament.ray_object_store[result_conf.id][result_instance] in tasks:
                        tasks.remove(result_tournament.ray_object_store[result_conf.id][result_instance])
            logger.info(f"Canceled TA: {result_conf.id}, {result_instance}")
        else:
            result_time = results[result_conf.id][result_instance]
            logger.info(f"TA result: {result_conf.id}, {result_instance} {result_time}")

        # Update the tournament based on result
        result_tournament, tournament_stop = tournament_dispatcher.update_tournament(results, tasks, result_conf,
                                                                                     result_tournament,
                                                                                     scenario.winners_per_tournament,
                                                                                     scenario.cutoff_time, scenario.par)

        logger.info(f"\nResult tournament update: {result_tournament}")

        if tournament_stop:
            tournament_counter += 1

            # Get the instances for the new tournament
            instance_id, instances = instance_selector.get_subset(result_tournament.instance_set_id + 1, time.time() - main_loop_start, result_tournament.instance_set_id + 1)
            all_configs = result_tournament.best_finisher + result_tournament.worst_finisher

            terminations = ray.get(global_cache.get_termination_history.remote())

            overall_best_update(tournaments, results, scenario)

            # Remove that old tournament
            tournaments.remove(result_tournament)
            tournament_history[result_tournament.id] = result_tournament
            global_cache.put_tournament_history.remote(result_tournament)

            ac_runtime = time.time() - main_loop_start
            logger.info(f"Results on instances: {results}")
            for surrogate in sm.surrogates.keys():
                start_update = time.time()
                if surrogate == Surrogates.GGApp:
                    if surrogate_update_counter == next_surrogate_update:
                        sm.update_surr(surrogate, result_tournament, all_configs, results, terminations, ac_runtime)

                        surrogate_update_counter = 0
                elif surrogate == Surrogates.SMAC:
                    sm.update_surr(surrogate, result_tournament, all_configs, results, terminations, ac_runtime)
                else:
                    sm.update_surr(surrogate, result_tournament, all_configs, results, terminations)
                surrogate_time = time.time() - start_update

            surrogate_update_counter = surrogate_update_counter + 1

            if surrogate_time >= surrogate_amortized_time:
                next_surrogate_update = next_surrogate_update + scenario.model_update_iteration
                surrogate_amortized_time = surrogate_amortized_time + surrogate_amortized_time

            # Generate and select
            random_points = [random_generator.point_generator() for _ in range(scenario.tournament_size * scenario.generator_multiple)]
            default_ps = [default_point_generator.point_generator()]

            hist = {**tournament_history, **{t.id: t for t in tournaments}}

            vg_points = [vg_point_generator.point_generator(
                         results=results, mode=Mode.random, alldata=hist,
                         lookback=1)
                         for _ in range(
                         scenario.tournament_size *
                         scenario.generator_multiple)]

            lhc_ps = lhc_point_generator.point_generator(
                n_samples=(scenario.tournament_size *
                           scenario.generator_multiple),
                lhs_type=LHSType.centered,
                criterion=Criterion.maximin)

            smac_conf = \
                sm.suggest(Surrogates.SMAC, scenario,
                           scenario.tournament_size * scenario.generator_multiple,
                           None, None, instances)

            ggapp_conf = \
                sm.suggest(Surrogates.GGApp, scenario,
                           scenario.tournament_size * scenario.generator_multiple,
                           hist, results, None)

            cppl_conf = \
                sm.suggest(Surrogates.CPPL, scenario,
                           scenario.tournament_size * scenario.generator_multiple,
                           None, None, instances)[0]
            
            generated_points = random_points + default_ps + \
                vg_points + lhc_ps + smac_conf + ggapp_conf + cppl_conf

            feature_time_start = time.time()

            features = \
                fg.static_feature_gen(generated_points, epoch, max_epochs)
            features = np.concatenate(
                (features, fg.diversity_feature_gen(generated_points, hist,
                                                    results, cutoff_time,
                                                    scenario.parameter,
                                                    predicted_quals,
                                                    evaluated)),
                axis=1)

            features = np.concatenate((features,
                                      fg.dynamic_feature_gen(generated_points,
                                                             hist,
                                                             None,
                                                             sm, cutoff_time,
                                                             results,
                                                             instances)),
                                      axis=1)
        
            feature_time = time.time() - feature_time_start

            set_weights = [value for hp, value in scenario.__dict__.items()
                           if hp[:2] == 'w_']
            weights = [set_weights for _ in generated_points]
            weights = np.array(weights)

            selection_start = time.time()

            points_to_run = \
                hp_seletor.select_points(scenario, generated_points,
                                         scenario.tournament_size - 1,
                                         epoch, max_epochs, features, weights,
                                         results, max_evals=100)

            select_time = time.time() - selection_start

            logger.info(f"All points generated \n\n{generated_points}\n\n")
            logger.info(f"Features computed \n\n{features.tolist()}\n\n")
            logger.info(f"Points selected to run \n\n{points_to_run}\n\n")
            logger.info(f"Instance set \n\n{instances}\n\n")
            logger.info(f"Terminations \n\n{terminations}\n\n")

            if scenario.wallclock_limit * inject_default < time.time() - main_loop_start:
                points_to_run[-1] = default_point_generator.point_generator()
                if tournament_counter % scenario.number_tournaments == 0:
                    inject_default += 0.25

            points_to_run = points_to_run + [result_tournament.best_finisher[0]]

            evaluated.extend(points_to_run)
            # Reduce evaluated list to tournament_size*tournament_size
            # after tn tournaments
            tn = 100
            if tournament_counter % tn == 0 and \
                    len(evaluated) > len(points_to_run):
                eval_np = []
                for ev_conf in evaluated:
                    eval_np.append(
                        sm.surrogates[Surrogates.GGApp].
                        transform_values(ev_conf))
                eval_np = np.asarray(eval_np)
                kmedoids = \
                    KMedoids(
                        n_clusters=scenario.number_tournaments * scenario.
                        tournament_size, random_state=0).fit(eval_np)
                clusters = kmedoids.cluster_centers_
                new_evaluated = []
                clusters = [c.tolist() for c in clusters]
                eval_np = [e.tolist() for e in eval_np]
                for c in clusters:
                    new_evaluated.append(evaluated[eval_np.index(c)])
                evaluated = new_evaluated

            pred_feature_time_start = time.time()

            for surrogate in sm.surrogates.keys():
                if surrogate is Surrogates.SMAC:
                    if sm.surrogates[surrogate].surr.model.rf is not None:
                        if qap:
                            predicted_quals.extend(sm.predict(surrogate,
                                                              points_to_run,
                                                              cutoff_time,
                                                              instances))
                        elif len(evaluated) != 0:
                            predicted_quals.extend(sm.predict(surrogate,
                                                              evaluated,
                                                              cutoff_time,
                                                              instances))
                            qap = True

                else:
                    if qap:
                        predicted_quals.extend(sm.predict(surrogate,
                                                          points_to_run,
                                                          cutoff_time,
                                                          instances))
                    elif len(evaluated) != 0:
                        predicted_quals.extend(sm.predict(surrogate,
                                                          evaluated,
                                                          cutoff_time,
                                                          instances))
                        qap = True

            pred_feat_time = time.time() - pred_feature_time_start

            # Create new tournament
            new_tournament, initial_assignments_new_tournament = tournament_dispatcher.init_tournament(results,
                                                                                                       points_to_run,
                                                                                                       instances,
                                                                                                       instance_id)

            # Add the new tournament and update the ray tasks with the new conf/instance assignments
            tournaments.append(new_tournament)
            tasks = update_tasks(tasks, initial_assignments_new_tournament, new_tournament, global_cache,  ta_wrapper, scenario)

            global_cache.put_tournament_update.remote(new_tournament)
            global_cache.remove_tournament.remote(result_tournament)
            # global_cache.put_tournament_update.remote(tournaments)

            logger.info(f"Final results tournament {result_tournament}")
            logger.info(f"New tournament {new_tournament}")
            epoch += 1
            
        else:
            # If the tournament does not terminate we get a new conf/instance assignment and add that as ray task
            next_task = tournament_dispatcher.next_tournament_run(results, result_tournament, result_conf)
            tasks = update_tasks(tasks, next_task, result_tournament, global_cache, ta_wrapper, scenario)
            logger.info(f"New Task {next_task}, {result_tournament}")
            global_cache.put_tournament_update.remote(result_tournament)

    global_cache.save_rt_results.remote()
    global_cache.save_tournament_history.remote()

    print('\n')
    print("AC run completed!")
    logger.info("AC run completed!")
    time.sleep(30)
    [ray.cancel(t) for t in not_ready]


if __name__ == "__main__":
    np.random.seed(42)

    parser = {"check_path": False, "seed": 42, "ta_run_type": "import_wrapper", "winners_per_tournament": 1, #import_wrapper
              "initial_instance_set_size": 2, "tournament_size": 2, "number_tournaments": 2, "total_tournament_number": 2,
              "total_runtime": 1200, "generator_multiple": 3, "set_size": 50, "monitor": "tournament_level",
              "termination_criterion": "total_tournament_number", "par": 1, "ta_pid_name": "glucose-simp", "memory_limit":1023*3, "log_folder":"run_1",
              "wrapper_mod_name": "selector.wrapper.cadical_wrapper", "wrapper_class_name": "Cadical_Wrapper"}

    scenario = Scenario("./selector/input/scenarios/test_example.txt", parser)#my_glucose_example #my_cadical_example

    check_log_folder(scenario.log_folder)
    clear_logs(scenario.log_folder)

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s', handlers=[
        logging.FileHandler(f"./selector/logs/{scenario.log_folder}/main.log"),
    ])

    logger = logging.getLogger(__name__)

    logger.info(f"Logging to {scenario.log_folder}")

    scenario = Scenario("./selector/input/scenarios/test_example.txt", parser)
    # TODO this needs to come from the scenario?!
    ta_wrapper = TAP_Work_Wrapper()

    # init
    ray.init(address="auto")
    # ray.init()

    logger.info("Ray info: {}".format(ray.cluster_resources()))
    logger.info("Ray nodes {}".format(ray.nodes()))
    logger.info("WD: {}".format(os.getcwd()))

    offline_mini_tournament_configuration(scenario, ta_wrapper, logger)

    save_latest_logs(scenario.log_folder)
    safe_best(sys.path[-1] + f'/selector/logs/{scenario.log_folder}/',
              scenario.cutoff_time)
    ray.shutdown()



"""This module contains surrogate management functions."""

import copy
from selector.pool import Surrogates
from selector.generators.surrogates.smac_surrogate import SmacSurr
from selector.generators.surrogates.ggapp_surrogate import GGAppSurr
from selector.generators.surrogates.cppl_surrogate import CPPL


class SurrogateManager():
    """Managing surrogates and related functions."""

    def __init__(self, scenario, seed=False, logger=None):
        """Initialize surrogate managing class.

        :param scenario: object, selector.scenario
        :param seed: int, random seed
        """
        self.seed = seed
        self.surrogates = {
            Surrogates.SMAC: SmacSurr(scenario, seed=self.seed,
                                      pca_dim=scenario.smac_pca_dim),
            Surrogates.GGApp: GGAppSurr(scenario, seed=self.seed, logger=logger),
            Surrogates.CPPL: CPPL(scenario, seed=self.seed,
                                  features=scenario.features)
        }

    def suggest(self, suggestor, scenario, n_samples, data, results,
                next_instance_set):
        """Suggest points based on surrogate.

        :param suggestor: object Surrogates, which surrogate to use
        :param scenario: object, selector.scenario
        :param n_samples: int, how many points to suggest
        :return sugg: list, suggested points
        """
        sugg = \
            self.surrogates[suggestor].get_suggestions(scenario,
                                                       n_samples,
                                                       data, results,
                                                       next_instance_set)

        return sugg

    def update_surr(self, surrogate, history, configs, results, terminations,
                    ac_runtime=None):
        """Update surrogate model with runhistory.

        :param surrogate: object Surrogates, which surrogate to use
        :param history: Tournament history
        :param conf: list, configurations, which history to update with
        :param state: object selector.pool.Status, status of this point
        :param tourn_nr: int, number of tournament, which to update with
        """
        confs = copy.deepcopy(configs)
        self.surrogates[surrogate].update(history, confs, results,
                                          terminations,
                                          ac_runtime=ac_runtime)

    def predict(self, surrogate, configs, cot, next_instance_set):
        """Get prediction for mean and variance concerning the points quality.

        :param surrogate: object Surrogates, which surrogate to use
        :param suggestions: list, suggested configurations
        :param cot: float, cut off time for tournaments
        :return predictions: list of dicts, contains info and predictions
            for regarded configurations
        """
        suggestions = copy.deepcopy(configs)
        if surrogate == Surrogates.SMAC:
            predict = self.surrogates[surrogate].predict(suggestions,
                                                         next_instance_set)

        try:
            predict = self.surrogates[surrogate].predict(suggestions,
                                                         next_instance_set)

            mean = predict[0]
            var = predict[1]

            return [{sugg.id: {'qual': mean[s], 'var': var[s],
                               'gen': sugg.generator}}
                    for s, sugg in enumerate(suggestions)]
        except:
            return [{sugg.id: {'qual': cot, 'var': 0,
                               'gen': sugg.generator}}
                    for sugg in suggestions]

    def ei(self, surrogate, suggestions, next_instance_set):
        """Compute expected improvement.

        :param surrogate: object Surrogates, which surrogate to use
        :param suggestions: list, suggested configurations
        :return ei: nested list, expected improvements
        """
        suggs = copy.deepcopy(suggestions)
        try:
            ei = self.surrogates[surrogate].\
                expected_improvement(suggs, next_instance_set)

            return ei
        except:
            return [[0] for sugg in suggestions]

    def pi(self, surrogate, suggestions, cot, results, next_instance_set):
        """Compute probability of improvement.

        :param surrogate: object Surrogates, which surrogate to use
        :param suggestions: list, suggested configurations
        :param cot: float, cut off time for tournaments
        :return pi: nested list, probabilities of improvement
        """
        suggs = copy.deepcopy(suggestions)
        pi = self.surrogates[surrogate].\
            probability_improvement(suggs, results, next_instance_set)

        return pi



import numpy as np
import json
import os
import ray
import pickle
import math
import copy
from selector.log_setup import ConfEncoder
from selector.generators.default_point_generator import check_conditionals

def get_conf_time_out(results, configuration_id, instances_set):
    """
    Determine if a configuration timed out on any instance in a set
    :param results:  dic. Run results of configuration
    :param configuration_id: int.
    :param instances_set: Set of instances.
    :return: None if no results are present, True if conf timed out on any instance, False if it dit not.
    """
    if configuration_id in list(results.keys()):
        conf_results_all_instances = results[configuration_id]
        conf_results_instances = [conf_results_all_instances[instance] for instance in
                                  conf_results_all_instances if instance in instances_set]
        if np.isnan(conf_results_instances).any():
            return True
        else:
            return False
    else:
        return None


def get_censored_runtime_for_instance_set(results, configuration_id, instances_set):
    """
    For a configuration compute the total runtime needed only for instances in a set. If there are no results for the
    conf return 0. Note that runs that were canceled by the monitor are not included since we count them as nan's
    :param results: Dic of results: {conf_id: {instance: runtime}}
    :param configuration_id: Id of the configuration
    :param instances_set: List of instances
    :return: Runtime of the configuration on instances
    """
    if configuration_id in results.keys():
        conf_results_all_instances = results[configuration_id]
        conf_results_instances = [conf_results_all_instances[instance] for instance in
                                  conf_results_all_instances if instance in instances_set]
        runtime = np.nansum(list(conf_results_instances))

    return runtime

def get_runtime_for_instance_set_with_timeout(results, configuration_id, instances_set, timeout, par_penalty=1):
    """
    For a configuration compute the total runtime needed only for instances in a set. If there are no results for the
    conf return 0. Note that runs that were canceled by the monitor are not included since we count them as nan's
    :param results: Dic of results: {conf_id: {instance: runtime}}
    :param configuration_id: Id of the configuration
    :param instances_set: List of instances
    :return: Runtime of the configuration on instances
    """
    if configuration_id in results.keys():
        conf_results_all_instances = results[configuration_id]
        conf_results_instances = [conf_results_all_instances[instance] for instance in
                                  conf_results_all_instances if instance in instances_set]

        runtime = np.nansum(list(conf_results_instances)) + np.count_nonzero(np.isnan(list(conf_results_instances))) * (timeout * par_penalty)

    return runtime


def get_censored_runtime_of_configuration(results, configuration_id):
    """
    Get total runtime of a conf not conditioned on an instance set. Note that runs that were canceled by the monitor
    are not included since we count them as nan's
    :param results: Dic of results: {conf_id: {instance: runtime}}
    :param configuration_id: Id of the configuration
    :return:
    """
    if configuration_id in results.keys():
        conf_results = results[configuration_id]
        runtime = np.nansum(list(conf_results.values()))
    return runtime

def get_instances_no_results(results, configuration_id, instance_set):
    """
    For a configuration get a list of instances we have no results for yet
    :param results: Dic of results: {conf_id: {instance: runtime}}
    :param configuration_id: Id of the configuration
    :param instance_set: List of instances
    :return: List of configuration the conf has not been run on
    """
    not_run_on= copy.deepcopy(instance_set)

    if configuration_id in results.keys():
        configuration_results = results[configuration_id]

        instances_run_on = configuration_results.keys()

        for iro in instances_run_on:
            if iro in not_run_on:
                not_run_on.remove(iro)

    return not_run_on

def overall_best_update(tournaments, results, scenario):
    """
    Over all tournaments get the best finisher with the most instance runs and shortest runtime and save that conf
    to a file.
    :param cache:
    :return:
    """

    number_of_instances_run = {}
    runtime = {}
    confs = {}
    # For each tournament get the best finisher
    for t in tournaments:
        if t.best_finisher:
            best_winner = t.best_finisher[0]
            number_of_instances_run[best_winner.id] = len(results[best_winner.id])
            #runtime[best_winner.id] = get_censored_runtime_of_configuration(results, best_winner.id)
            runtime[best_winner.id] = get_runtime_for_instance_set_with_timeout(results, best_winner.id, t.instance_set, scenario.cutoff_time, scenario.par)
            confs[best_winner.id] = best_winner
    # If we have any best finisher we get those which ran on the most instances and then get the conf with the
    # shortest runtime on those
    if confs:
        max_number_instances = max(list(number_of_instances_run.values()))
        conf_with_max_number = [k for k, v in number_of_instances_run.items() if v == max_number_instances]

        conf_r_w_max_i = {k: runtime[k] for k in conf_with_max_number}
        conf_with_min_runtime = min(list(conf_r_w_max_i.values()))
        conf_with_min_runtime = [k for k, v in conf_r_w_max_i.items() if v == conf_with_min_runtime]


        clean_conf = copy.copy(confs[conf_with_min_runtime[0]].conf)
        # Check conditionals and turn off parameters if violated
        cond_vio = check_conditionals(scenario, clean_conf)
        for cv in cond_vio:
            clean_conf.pop(cv, None)

        with open(f"./selector/logs/{scenario.log_folder}/trajectory.json", 'a') as f:
            json.dump({str(confs[conf_with_min_runtime[0]].id): clean_conf}, f, cls=ConfEncoder)
            f.write(os.linesep)




def config_translator(s, config):
    """
    Converts boolean values back to their original representation
    :param s: scenario class object
    :param config: configuration class object
    :return: dic. parameter configuration with converted boolean values
    """

    boolean1 = ["on", "off"]
    boolean2 = ["yes", "no"]

    for param in s.parameter:
        if not param.original_bound == []:

            if param.original_bound == boolean1:
                if config.conf[param.name]:
                    config.conf[param.name] = "on"
                else:
                    config.conf[param.name] = "off"

            if param.original_bound == boolean2:
                if config.conf[param.name]:
                    config.conf[param.name] = "yes"
                else:
                    config.conf[param.name] = "no"

    return config




import logging
import time
import subprocess
import ray
import numpy as np
# import psutil
import os
import signal
import traceback
import re
import copy
import sys
from threading import Thread
from queue import Queue, Empty

sys.stderr = open(os.devnull, "w")
try:
    import psutil
finally:
    sys.stderr = sys.stderr

def enqueue_output(out, queue):
    for line in iter(out.readline, b''):
        line = line.decode("utf-8")
        queue.put(line)
    out.close()

def get_running_processes(ta_process_name):
    processes = []
    for proc in psutil.process_iter():
        try:
            processName = proc.name()
            processID = proc.pid
            if processName in [ta_process_name]:
                processes.append([processName, processID])
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    return processes

def termination_check(process_pid, process_status, ta_process_name, python_pid, conf_id, instance):
    running_processes = get_running_processes(ta_process_name)

    sr = False
    for rp in running_processes:
        if process_pid == rp[1]:
            sr = True

    if sr:
        logging.info(f"Failed to terminate {conf_id}, {instance}: process {process_pid} with {process_status} on {python_pid} is still running")
    else:
        logging.info(
            f"Successfully terminated {conf_id}, {instance} on {python_pid} with {process_status}")

def tae_from_cmd_wrapper_rt():
    """
    Execute the target algorithm with a given conf/instance pair by calling a user provided Wrapper that created a cmd
    line argument that can be executed
    :param conf: Configuration
    :param instance: Instances
    :param cache: Cache
    :param ta_command_creator: Wrapper that creates a
    :return:
    """
    # todo logging dic should be provided somewhere else -> DOTAC-37

    try:
        cmd = 'stdbuf -oL /media/dweiss/Transcend5/AC_architecture/Selector/selector/cadical/cadical  --arena=1 --arenacompact=true --arenasort=1 --binary=false --check=false --compact=false --compactint=43502 --compactmin=8601 --elim=false --elimclslim=45931 --elimint=569853 --elimocclim=427 --elimrounds=141 -q /media/dweiss/Transcend5/AC_architecture/Selector/selector/cadical/circuit_fuzz/fuzz_100_28405.cnf'

        p = psutil.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT, close_fds=True)

        print(p.poll())

        q = Queue()
        t = Thread(target=enqueue_output, args=(p.stdout, q))
        t.daemon = True
        t.start()
        start = time.time()

        timeout = False
        empty_line = False
        memory_p = 0
        cpu_time_p = 0
        reading = True

        while reading:
            try:
                line = q.get(timeout=.5)
                empty_line = False
                print('\ncpu_time_p', cpu_time_p)
                # Get the cpu time and memory of the process
            except Empty:
                empty_line = True
                if p.poll() is None:
                    #cpu_time_p = time.time() - start
                    #print(cpu_time_p)
                    cpu_time_p = p.cpu_times().user
                    print('cpu time 109', cpu_time_p, p.cpu_times())
                    # memory_p = p.memory_info().rss / 1024 ** 2
                    # print(p.memory_info().rss)
                if float(cpu_time_p) > float(10) or float(memory_p) > float(10) and timeout ==False:
                    timeout = True
                    if p.poll() is None:
                        p.terminate()
                    try:
                        time.sleep(1)
                    except:
                        pass
                    if p.poll() is None:
                        p.kill()
                    try:
                        os.killpg(p.pid, signal.SIGKILL)
                    except Exception:
                        pass
                    # if scenario.ta_pid_name is not None:
                    #    termination_check(p.pid, p.poll(), scenario.ta_pid_name, os.getpid(),conf.id, instance_path)
                pass
            else:  # write intemediate feedback
                print(line)

            if p.poll() is None:
                # Get the cpu time and memory of the process
                try:
                    cpu_time_p = time.time() - start
                    print(cpu_time_p)
                    # cpu_time_p = p.cpu_times().user
                    # memory_p = p.memory_info().rss / 1024 ** 2
                except Exception as e:
                    print("Looking for this", e, p.poll(), cpu_time_p)
                    pass

                if float(cpu_time_p) > float(10) or float(memory_p) > float(
                        10) and timeout == False:
                    timeout = True
                    if p.poll() is None:
                        p.terminate()
                    try:
                        time.sleep(1)
                    except:
                        pass
                    if p.poll() is None:
                        p.kill()
                    try:
                        os.killpg(p.pid, signal.SIGKILL)
                    except Exception:
                        pass
                    # if scenario.ta_pid_name is not None:
                    #    termination_check(p.pid, p.poll(), scenario.ta_pid_name, os.getpid(),conf.id, instance_path)

            # Break the while loop when the ta was killed or finished
            if empty_line and p.poll() != None:
                reading = False

        time.sleep(0.2)

    except Exception:
        print({traceback.format_exc()})
        logging.info(f"Exception in TA execution: {traceback.format_exc()}")

if __name__ == "__main__":

    tae_from_cmd_wrapper_rt()



u"""This module contains the variable graph point generator.

Based on the variable tree in GGA [Ansótegui, C., Sellmann, M., Tierney, K.:
A Gender-Based Genetic Algorithm for the Automatic Configuration of Algorithms.
In: Principles and Practice of Constraint Programming - CP 2009.
pp. 142–157 (09 2009).]
"""

import pickle
import os
import random
import copy
import numpy as np
import itertools
import math
from enum import Enum, IntEnum
from selector.pool import Configuration, ParamType, Generator
from selector.generators.random_point_generator import (
    random_set_conf,
    reset_conditionals
)
from selector.generators.default_point_generator import (
    check_conditionals,
    check_no_goods
)


class LabelType(IntEnum):
    """Contains the types of labels a node can assume."""

    C = 1
    N = 2
    O = 3


class Mode(Enum):
    """Contains the types of modes for choosing parent configurations."""

    random = 1
    only_best = 2
    best_and_random = 3


def variable_graph_structure(s):
    """
    General variable graph structure is read from scenario.

    : param s: scenario object
    return: general variable graph structure
    """
    parameters = [param.name for param in s.parameter]
    graph_structure = {}
    for p in parameters:
        graph_structure[p] = []

    for pi, pj in zip(s.parameter[:-1], s.parameter[1:]):
        if pi.name not in graph_structure[pi.name]:
            graph_structure[pi.name] = []
        graph_structure[pi.name].append(pj.name)
        if pi.condition:
            for cond in pi.condition:
                if cond not in graph_structure[pi.name]:
                    graph_structure[pi.name].append(cond)

    for ng in s.no_goods:
        for param in ng:
            for ngp, _ in ng.items():
                if ngp not in graph_structure[param] and ngp != param:
                    graph_structure[param].append(ngp)

    return graph_structure


def decide_for_O(config_label, C, N, cn):
    """
    Decide new label in case of label O.

    : param config_label: current labeling of parameters
    : param C: first configuration
    : param N: second configuration
    : param cn: child node in the graph
    return: config_label
    """
    if config_label[cn] == LabelType.O:
        if cn in C.conf and cn in N.conf:
            config_label[cn] = LabelType.C
            if C.conf[cn] != N.conf[cn] and random.uniform(0, 1) >= 0.5:
                config_label[cn] = LabelType.N
        elif cn in C.conf and cn not in N.conf:
            if random.uniform(0, 1) >= 0.5:
                config_label[cn] = LabelType.C
            else:
                del config_label[cn]
        elif cn not in C.conf and cn in N.conf:
            if random.uniform(0, 1) >= 0.5:
                config_label[cn] = LabelType.N
            else:
                del config_label[cn]
        else:
            del config_label[cn]

    return config_label


def check_valid(config_label, C, N, cn):
    """
    Check if parameter is actually set in Configuration.

    : param config_label: current labeling of parameters
    : param C: first configuration
    : param N: second configuration
    return: config_label
    """
    if cn in config_label:
        if config_label[cn] == LabelType.C and cn not in C.conf:
            config_label.pop(cn, None)
        elif config_label[cn] == LabelType.N and cn not in N.conf:
            config_label.pop(cn, None)

    return config_label


def set_config_label(paths, config_label, cn, C, N):
    """
    Set label.

    : param paths: nodes visited until current node
    : param config_label: current labeling
    : param cn: current node
    : param C: first Configuration
    : param N: second configuration
    return: config_label
    """
    parent_nodes = copy.copy(paths[cn])
    parent_nodes.remove(cn)
    parent_labels = [config_label[x] for x in parent_nodes
                     if x in config_label]
    values, counts = np.unique(parent_labels,
                               return_counts=True)
    config_label[cn] = random.choices(values, weights=counts,
                                      k=1)[0]
    config_label = decide_for_O(config_label, C, N, cn)
    config_label = check_valid(config_label, C, N, cn)

    return config_label


def reset_no_goods(s, config_setting, label, C, N):
    """
    Check if no goods violated and change value if so.

    : param s: scenario
    : param config_setting: parameter value setting
    : param label: label
    : param C: first Configuration
    : param N: second configuration
    return: config_setting
    """
    for ng in s.no_goods:
        params = list(ng.keys())
        if all(ng[p] == config_setting[p] for p in params):
            for p in params:
                if label[p] == LabelType.C and p in N.conf \
                        and N.conf[p] != ng[p]:
                    config_setting[p] = N.conf[p]
                elif label[p] == LabelType.N and p in C.conf \
                        and C.conf[p] != ng[p]:
                    config_setting[p] = C.conf[p]
                else:
                    for param in s.parameter:
                        if p == param.name:
                            config_setting[p] = \
                                random_set_conf([param])[p]

    return config_setting


def graph_crossover(graph_structure, C, N, s):
    """
    Crossover according to variable graph.

    : param graph_structure: general variable graph structure
    : param C: configuration 1
    : param N: configuration 2
    : param s: scenario
    return: new configuration setting
    """
    params = list(graph_structure.keys())
    curr_node = params[0]

    config_label = {}
    config_setting = {}
    paths = {}
    S = [curr_node]

    # Label root node
    if C.conf[curr_node] == N.conf[curr_node] or\
            len(graph_structure[curr_node]) > 1:
        config_label[curr_node] = LabelType.O
    else:
        config_label[curr_node] = random.choice([LabelType.C, LabelType.N])

    paths[curr_node] = [curr_node]

    while S:
        curr_node = S[0]
        S.pop(0)
        child_nodes = graph_structure[curr_node]
        for cn in child_nodes:
            if curr_node != cn:
                if cn in paths:
                    if cn not in paths[cn]:
                        paths[cn].append(cn)
                        config_label = set_config_label(paths,
                                                        config_label,
                                                        cn, C, N)
                        S.append(cn)
                else:
                    paths[cn] = [*paths[curr_node], cn]
                    config_label = set_config_label(paths, config_label,
                                                    cn, C, N)
                    S.append(cn)

                if random.uniform(0, 1) < 0.1:
                    if cn in config_label:
                        if config_label[cn] == LabelType.N and cn in C.conf:
                            config_label[cn] = LabelType.C
                        elif config_label[cn] == LabelType.C and cn in N.conf:
                            config_label[cn] = LabelType.N
                    S.append(cn)

    for param, label in config_label.items():
        config_setting[param] = N.conf[param] if label == LabelType.N \
            else C.conf[param]

    param_info = {}
    for param in s.parameter:
        param_info[param.name] = {'type': param.type, 'bound': param.bound}

    for param, value in config_setting.items():
        if param_info[param]['type'] == ParamType.categorical:
            if random.uniform(0, 1) < 0.1:
                config_setting[param] = \
                    random.choice(param_info[param]['bound'])
        else:
            if random.uniform(0, 1) < 0.1:
                mutation = param_info[param]['bound'][0] - 1
                while mutation < param_info[param]['bound'][0] or \
                        mutation > param_info[param]['bound'][1]:
                    mu = config_setting[param]
                    sigma = (param_info[param]['bound'][1] -
                             param_info[param]['bound'][0]) * 0.1
                    mutation = \
                        np.random.normal(mu, sigma, 1)[0]
                if param_info[param]['type'] == ParamType.continuous:
                    config_setting[param] = mutation
                else:
                    config_setting[param] = round(mutation)

    # Check no goods and reset values if violated
    ng_vio = check_no_goods(s, config_setting)
    while ng_vio:
        config_setting = reset_no_goods(s, config_setting,
                                        config_label, C, N)
        ng_vio = check_no_goods(s, config_setting)

    return config_setting


def choose_parents(mode, data, lookback, results):
    """
    Pick Configurations according to mode.

    : param mode: Enum, mode of parent selection
    : param data: Tournament data to select parents from
    : param lookback: int, how many tournaments from the past are included
    return: configurations C and N
    """
    if lookback < len(data):
        data = list(data.values())[:len(data) - lookback]
    else:
        data = list(data.values())

    if mode == Mode.random:
        conf_list = []

        for tourn in data:
            conf_list.extend(tourn.best_finisher)
            conf_list.extend(tourn.worst_finisher)

        if conf_list:
            C = np.random.choice(conf_list)
            C_ind = conf_list.index(C)
            conf_list.pop(C_ind)
        else:
            C = np.random.choice(tourn.configurations)

        if conf_list:
            N = np.random.choice(conf_list)
        else:
            N = np.random.choice(tourn.configurations)

    elif mode == Mode.only_best:
        all_best = []
        for tourn in data:
            all_best.append(tourn.best_finisher[0])

        if len(all_best) > 1:
            C = np.random.choice(all_best)
            best_ind = all_best.index(C)
            all_best.pop(best_ind)
            N = np.random.choice(all_best)

        else:
            mode = Mode.best_and_random

    elif mode == Mode.best_and_random:
        all_best = []
        all_worst = []
        for tourn in data:
            all_best.extend(tourn.best_finisher)

        performance = []
        for ab in all_best:
            perf = 0
            for res in results[ab.id].values():
                perf += res
            perf = perf / len(results[ab.id])
            performance.append(perf)

        take_best = math.ceil(len(all_best) * 0.1)
        sort_index = np.argsort(performance)
        the_best = sort_index[:take_best]
        best = []
        for tb in the_best:
            best.append(all_best[tb])

        all_best = best
        for tourn in data:
            for tbf in tourn.best_finisher:
                if tbf in all_best:
                    continue
                else:
                    all_worst.append(tbf)
            all_worst.extend(tourn.worst_finisher)

        all_configs = []

        if all_best:
            C = np.random.choice(all_best)
        else:
            for tourn in data:
                all_configs.extend(tourn.configurations)
            C = np.random.choice(all_configs)

        if all_worst:
            N = np.random.choice(all_worst)
        else:
            if all_configs:
                pass
            else:
                for tourn in data:
                    all_configs.extend(tourn.configurations)
            N = None
            while N != C:
                N = np.random.choice(all_configs)

    return C, N


def variable_graph_point(s, identity, results, mode=Mode.best_and_random,
                         alldata=False, lookback=1, seed=False):
    """
    Configuration is generated via variable graph method.

    : param s: scenario object
    : param identity: uuid to identify configuration
    : param mode: Enum, mode of parent selection
    : param data: Tournament data to select parents from
    : param lookback: int, how many tournaments from the past are included
    : param seed: sets random seed
    return: configuration
    """
    if seed:
        np.random.seed(seed)

    if not alldata:
        print('No data given to variable point generator')
        exit()
    # Pick parent configurations
    data = copy.copy(alldata)
    C, N = choose_parents(mode, data, lookback, results)

    # Generate general graph structure
    graph_structure = variable_graph_structure(s)

    # Generate configuration via variable graph crossover
    config_setting = graph_crossover(graph_structure, C, N, s)

    # Fill Configuration class with ID and parameter values
    configuration = Configuration(identity,
                                  config_setting,
                                  Generator.var_graph)

    return configuration



import ray
import logging
import time
import numpy as np


from selector.tournament_performance import get_censored_runtime_for_instance_set,get_conf_time_out, get_runtime_for_instance_set_with_timeout


@ray.remote(num_cpus=1)
class Monitor:
    def __init__(self, sleep_time, cache, scenario):
        """
        Monitor whether the live total runtime of a running conf is exceeding the accumulated runtime of the worst finisher,
         given that we have already enough finisher. While up the monitor may kill multiple conf/instance pairs. To avoid
         killing a ta run twice, the monitor stores what it has already killed.
        :param sleep_time: Int. Wake up and check whether runtime is exceeded
        :param cache: Ray cache
        :param number_of_finisher: Int.
        :return: conf/instance that are killed.
        """
        self.sleep_time = sleep_time
        self.cache = cache
        self.number_of_finisher = scenario.winners_per_tournament
        self.tournaments = []
        self.time_out = scenario.cutoff_time
        self.par = scenario.par

        logging.basicConfig(filename=f'./selector/logs/{scenario.log_folder}/monitor.log', level=logging.INFO,
                            format='%(asctime)s %(message)s')

    def monitor(self):
        logging.info("Starting monitor")
        while True:
            # Get results that are already available for ta runs
            start = time.time()
            results = ray.get(self.cache.get_results.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting results {dur}")

            # get starting times for each conf/instance
            start = time.time()
            start_time = ray.get(self.cache.get_start.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting start {dur}")

            # Get the current tournaments that are in the cache
            start = time.time()
            tournaments = ray.get(self.cache.get_tournament.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting tournaments {dur}")

            start = time.time()
            termination_history = ray.get(self.cache.get_termination_history.remote())
            dur = time.time() - start
            logging.info(f"Monitor getting termination_history {dur}")

            for t in tournaments:
                # We can only start canceling runs if there are enough winners already
                if len(t.best_finisher) == self.number_of_finisher:
                    # Compare runtime to the worst best finisher
                    worst_best_finisher = t.best_finisher[-1]
                    runtime_worst_best_finisher = get_runtime_for_instance_set_with_timeout(results, worst_best_finisher.id,
                                                                                        t.instance_set, self.time_out, self.par)
                    # We need to compare each configuration that is still in the running to the worst finisher
                    for conf in t.configurations:
                        # Here we figured out which instances the conf is still running and which one it already finished
                        if conf.id in list(results.keys()):
                            instances_conf_finished = list(results[conf.id].keys())
                            conf_runtime_f = get_runtime_for_instance_set_with_timeout(results, conf.id, t.instance_set, self.time_out, self.par)
                        else:
                            instances_conf_finished = []
                            conf_runtime_f = 0
                        instances_conf_planned = list(t.ray_object_store[conf.id].keys())
                        instances_conf_still_runs = [c for c in instances_conf_planned if c not in instances_conf_finished]

                        # The runtime of a conf is the time it took to finish instances plus the time spend running but
                        # not finishing the running instances
                        # if i in list(start_time[conf.id].keys()): is a bit hack: it might be the case that the main
                        # process things a conf/instance pair is running but the cache has not recived a start time and
                        # thus the conf/instance is in a transition. That conf instance then has not runtime yet
                        # (or at least very very little) so we ignore it for the cancel computation
                        conf_runtime_p = sum([(time.time() - start_time[conf.id][i]) for i in instances_conf_still_runs if i in list(start_time[conf.id].keys())])
                        conf_runtime = conf_runtime_f + conf_runtime_p
                        #conf_time_out = get_conf_time_out(results, conf.id, t.instance_set)

                        logging.info(f"Monitor kill check,{conf.id} {conf_runtime}, {runtime_worst_best_finisher}"
                                     f"{worst_best_finisher.id,}, {[m. id for m in t.configurations]}")

                        if conf_runtime > runtime_worst_best_finisher:# or conf_time_out:
                            # We can only kill still running tasks
                            for i in instances_conf_still_runs:
                                # We check if we have killed the conf/instance pair before.
                                if self.termination_check(conf.id, i, termination_history):
                                    logging.info(f"Monitor is killing: {conf} {i} with id: {t.ray_object_store[conf.id][i]}")
                                    print(f"Monitor is killing:{time.ctime()} {t.ray_object_store[conf.id][i]}")
                                    # In case we kill we store that we have killed
                                    self.cache.put_termination_history.remote(conf.id, i)
                                    [ray.cancel(t.ray_object_store[conf.id][i])]
                                else:
                                    continue
            time.sleep(self.sleep_time)

    def termination_check(self, conf_id, instance, termination_history):
        """
        Check if we have killed a conf/instance pair already. Return True if we did not.
        :param conf_id:
        :param instance:
        :return:
        """
        if conf_id not in termination_history:
            return True
        elif instance not in termination_history[conf_id]:
            return True
        else:
            return False










"""This module contains feature generation functions."""
import numpy as np
import copy
import time
from collections import defaultdict
from selector.pool import Generator, Surrogates
import selector.hp_point_selection as hps


class FeatureGenerator:
    """Generate features necessary to evaluate configurations."""

    def __init__(self, logger=None):
        """Initialize feature generation class."""
        self.Generator = Generator
        self.logger = logger

    def percent_rel_evals(self, suggestions, data, nce):
        """Percentage of relatives so far evaluated.

        :param suggestions: list, suggested points
        :param data: data object, contains historic performance data
        :param nce: int, number of configuration evaluations
        :return div_feats: list, computed features of suggested points
        """
        self.gen_counts = defaultdict(int)

        for content in data.values():
            for finisher in content.best_finisher + content.worst_finisher:
                self.gen_counts[finisher.generator] += 1

        div_feats = [[self.gen_counts[sugg.generator]]
                     for sugg in suggestions]

        max_val = max(max(d) for d in div_feats)
        if max_val != 0.0:
            for d in div_feats:
                d[0] = d[0] / max_val

        return div_feats

    def avg_rel_evals_qual(self, suggestions, data, nce, results, cot,
                           generators):
        """Average quality of relatives so far evaluated.

        :param suggestions: list, suggested points
        :param data: data object, contains historic performance data
        :param nce: int, number of all configs evaluated
        :param results: dict, qualities of configurations
        :param cot: float, cut off time for tournaments
        :param generators: list, all possible generators
        :return div_feats: list, computed features of suggested points
        """
        div_feats = []
        quals = {}
        gen_counts = self.gen_counts
        quals = defaultdict(int)
        for content in data.values():
            for finisher in content.best_finisher + content.worst_finisher:
                quals[finisher.generator] += \
                    sum(results[finisher.id].values()) / \
                    len(results[finisher.id])

        for gen in quals.keys():
            quals[gen] /= gen_counts.get(gen, 1)  # Avoid division by zero

        div_feats = [[quals.get(sugg.generator, 0) / cot]
                     for sugg in suggestions]

        max_val = max(max(d) for d in div_feats)
        if max_val != 0.0:
            for d in div_feats:
                d[0] /= max_val

        return div_feats

    def best_rel_evals_qual(self, suggestions, data, generators, results, cot):
        """Best target value relatives so far evaluated.

        :param suggestions: list, suggested points
        :param data: data object, contains historic performance data
        :param generators: list, all possible generators
        :param results: dict, qualities of configurations
        :param cot: float, cut off time for tournaments
        :return div_feats: list, computed features of suggested points
        """
        div_feats = []
        best_val = {}
        best_val = defaultdict(int)
        for content in data.values():
            for finisher in content.best_finisher + content.worst_finisher:
                for val in results[finisher.id].values():
                    if finisher.generator not in best_val:
                        best_val[finisher.generator] = val
                    elif val < best_val[finisher.generator]:
                        best_val[finisher.generator] = val

        div_feats = [[best_val.get(sugg.generator, 0) / cot]
                     for sugg in suggestions]

        max_val = max(max(d) for d in div_feats)
        if max_val != 0.0:
            for d in div_feats:
                d[0] = d[0] / max_val

        return div_feats

    def std_rel_evals_qual(self, suggestions, data, generators, results, cot):
        """Std of quality of relatives so far evaluated.

        :param suggestions: list, suggested points
        :param data: data object, contains historic performance data
        :param generators: list, all possible generators
        :param results: dict, qualities of configurations
        :param cot: float, cut off time for tournaments
        :return div_feats: list, computed features of suggested points
        """
        div_feats = []
        qual_vals = {}
        qual_vals = defaultdict(int)
        for content in data.values():
            for finisher in content.best_finisher + content.worst_finisher:
                if finisher.generator in qual_vals:
                    for res_val in list(results[finisher.id].values()):
                        qual_vals[finisher.generator].append(res_val)
                else:
                    qual_vals[finisher.generator] = \
                        list(results[finisher.id].values())

        qual_std = {}

        for key, qv in qual_vals.items():
            qual_std[key] = np.std(qv)

        div_feats = [[qual_std.get(sugg.generator, 0) / cot]
                     for sugg in suggestions]

        max_val = max(max(d) for d in div_feats)
        if max_val != 0.0:
            for d in div_feats:
                d[0] = d[0] / max_val

        return div_feats

    def diff_pred_real_qual(self, suggestions, data, predicted_quals, results):
        """Difference of predicted & real qual. of relatives evaluated so far.

        :param suggestions: list, suggested points
        :param data: data object, contains historic performance data
        :param predicted_quals: nested list, predicted performance/quality for
            suggested configurations
        :param results: dict, qualities of configurations
        :return div_feats: list, computed features of suggested points
        """
        if not predicted_quals:
            div_feats = [[0] for _ in suggestions]
        else:
            rel_results = {}
            rel_predicts = {}
            div_feats = []
            diffs = {}

            # Collect relevant predictions into a dictionary
            for pred in predicted_quals:
                pred = list(pred.values())[0]
                gen = pred['gen']
                rel_predicts.setdefault(gen, []).append(pred['qual'])

            # Collect relevant results into a dictionary
            for content in data.values():
                for finisher in content.best_finisher + content.worst_finisher:
                    gen = finisher.generator
                    rel_results.setdefault(gen, []).extend(results[finisher.id].values())

            for gen in Generator:
                if gen in rel_results and gen in rel_predicts:
                    if len(rel_predicts[gen]) > 0 and \
                            len(rel_results[gen]) > 0:
                        diffs[gen] = \
                            np.mean(rel_predicts[gen]) \
                            / np.mean(rel_results[gen])
                elif gen not in diffs:
                    diffs[gen] = 0

            for sugg in suggestions:
                div_feats.append([diffs[sugg.generator]])

            max_val = max(max(d) for d in div_feats)
            if max_val != 0.0:
                for d in div_feats:
                    d[0] = d[0] / max_val

        return div_feats

    def avg_dist_evals(self, suggests, evals, psetting):
        """Average distance to all points so far evaluated.

        :param suggestions: list, suggested points
        :param evals: list, already evaluated points
        :param psetting: scenario.parameter
        :return div_feats: list, average distances to all already evaluated
            points
        """
        if evals:

            suggestions = copy.deepcopy(suggests)
            evaluated = copy.deepcopy(evals)

            suggestions = hps.normalize_plus_cond_acc(suggestions, psetting)
            evaluated = hps.normalize_plus_cond_acc(evaluated, psetting)
            distances = hps.pairwise_distances(suggestions, evaluated)

            div_feats = []
            for dist in distances:
                div_feats.append([np.mean(dist)])

            max_val = max(max(d) for d in div_feats)
            if max_val != 0.0:
                for d in div_feats:
                    d[0] = d[0] / max_val

        else:
            div_feats = [[0] for _ in suggests]

        return div_feats

    def avg_dist_sel(self, suggests, psetting):
        """Average distance to points in the current selection.

        :param suggestions: list, suggested points
        :param psetting: scenario.parameter
        :return div_feats: list, average distances to points in current
                                 selection
        """
        suggestions = copy.deepcopy(suggests)

        suggestions = hps.normalize_plus_cond_acc(suggestions, psetting)
        distances = hps.pairwise_distances(suggestions, suggestions)

        div_feats = []
        for dist in distances:
            div_feats.append([np.mean(dist)])

        max_val = max(max(d) for d in div_feats)
        if max_val != 0.0:
            for d in div_feats:
                d[0] = d[0] / max_val

        return div_feats

    def avg_dist_rel(self, suggests, evals, psetting, generators):
        """Average distances to relatives.

        :param suggests: list, suggested points
        :param evals: list, already evaluated points
        :param psetting: scenario.parameter
        :param generators: list, available generators
        :return div_feats: list, computed features of suggested points
        """
        if evals:
            suggestions = copy.deepcopy(suggests)
            evaluated = copy.deepcopy(evals)

            suggestions = hps.normalize_plus_cond_acc(suggestions, psetting)
            evaluated = hps.normalize_plus_cond_acc(evaluated, psetting)

            group_relatives = {}
            for gen in generators:
                for ev in evaluated:
                    if gen == ev.generator:
                        if gen not in group_relatives:
                            group_relatives[gen] = [ev]
                        else:
                            group_relatives[gen].append(ev)

            distances = []
            for sugg in suggestions:
                if sugg.generator in group_relatives:
                    distances.append(hps.pairwise_distances([sugg],
                                     group_relatives[sugg.generator]))
                else:
                    distances.append([0 for _ in sugg.conf])

            div_feats = []
            for dist in distances:
                div_feats.append([np.mean(dist)])

            max_val = float(max(max(d) for d in div_feats))
            for d in div_feats:
                if max_val != 0.0:
                    d[0] = d[0] / max_val

        else:
            div_feats = [[0] for _ in suggests]

        return div_feats

    def expected_qual(self, suggs, sm, cot, surr, next_instance_set):
        """Expected quality of points.

        :param suggests: list, suggested points
        :param sm: object, surrogates.SurrogateManager()
        :param cot: int, cut off time
        :param surr: which surrogate to use
        :param next_instance_set: next instances that will be run
        :return dyn_feats: list, computed features of suggested points
        """
        suggests = copy.deepcopy(suggs)
        dyn_feats = []
        try:
            expimp = sm.predict(surr, suggests, cot, next_instance_set)
            self.expimp = expimp

            for exim in expimp:
                for ei in exim.values():
                    dyn_feats.append([ei['qual']])

            max_val = float(max(max(d) for d in dyn_feats))
            for d in dyn_feats:
                if max_val != 0.0:
                    d[0] = d[0] / max_val

        except:
            dyn_feats = [[0] for _ in suggests]

        return dyn_feats

    def prob_qual_improve(self, suggs, sm, cot, results, surr,
                          next_instance_set):
        """Probability of quality of points to improve.

        :param suggests: list, suggested points
        :param sm: object, surrogates.SurrogateManager()
        :param cot: int, cut off time
        :param results: list, results of points evaluated so far
        :param surr: which surrogate to use
        :return dyn_feats: list, computed features of suggested points
        """
        suggests = copy.deepcopy(suggs)
        dyn_feats = []
        try:
            expimp = sm.pi(surr, suggests, cot, results, next_instance_set)

            for ei in expimp:
                dyn_feats.append(list(ei))

            max_val = float(max(max(d) for d in dyn_feats))
            for d in dyn_feats:
                if max_val != 0.0:
                    d[0] = d[0] / max_val

        except:
            dyn_feats = [[0] for _ in suggests]

        return dyn_feats

    def uncertainty_improve(self, suggs, sm, cot, surr, next_instance_set):
        """Probability of quality of points to improve.

        :param suggests: list, suggested points
        :param sm: object, surrogates.SurrogateManager()
        :param cot: int, cut off time
        :param surr: which surrogate to use
        :return dyn_feats: list, computed features of suggested points
        """
        suggests = copy.deepcopy(suggs)
        dyn_feats = []
        try:
            expimp = self.expimp

            for exim in expimp:
                for ei in exim.values():
                    dyn_feats.append([ei['var']])

            max_val = float(max(max(d) for d in dyn_feats))
            for d in dyn_feats:
                if max_val != 0.0:
                    d[0] = d[0] / max_val

        except:
            dyn_feats = [[0] for _ in suggests]

        return dyn_feats

    def expected_improve(self, suggs, sm, cot, surr, next_instance_set):
        """Probability of quality of points to improve.

        :param suggests: list, suggested points
        :param sm: object, surrogates.SurrogateManager()
        :param cot: int, cut off time
        :param surr: which surrogate to use
        :return dyn_feats: list, computed features of suggested points
        """
        suggests = copy.deepcopy(suggs)
        dyn_feats = []
        try:
            expimp = sm.ei(surr, suggests, next_instance_set)
            for ei in expimp:
                dyn_feats.append([ei])

            max_val = float(max(max(d) for d in dyn_feats))
            for d in dyn_feats:
                if max_val != 0.0:
                    d[0] = d[0] / max_val

        except:
            dyn_feats = [[0] for _ in suggests]

        return dyn_feats

    def surr_votes(self, dyn_feats):
        """Multiply surr features to get agreement features.

        :param dyn_feats: list of np.ndarray, dynamic features.
        :return dyn_feats: list of np.ndarray, extended dynamic features.
        """
        nr_surrs = len(Surrogates)
        new_feature_sets = []

        for k in range(0, len(dyn_feats[0]), nr_surrs):
            votes = [[] for _ in range(nr_surrs)]
            for i, _ in enumerate(dyn_feats):
                for j in range(nr_surrs):
                    votes[j].append([dyn_feats[i, k] * dyn_feats[i, k - 1]])
            new_feature_sets.append(votes)

        for nfs in new_feature_sets:
            for v in nfs:
                dyn_feats = np.concatenate((dyn_feats, v), axis=1)

        return dyn_feats

    def static_feature_gen(self, suggestions, epoch, max_epoch):
        """Generate static features.

        :param suggestions: list, suggested configurations
        :param epoch: int, current epoch
        :param max_epoch: int, total number of epochs
        :return static_features: list, static features
        """
        if self.logger is not None:
            static_time = time.time()

        static_feats = [[] for ii in range(len(suggestions))]

        # One-Hot encoded information of generator used for conf
        for s in range(len(suggestions)):
            for gt in range(len(self.Generator)):
                if suggestions[s].generator == self.Generator(gt + 1):
                    static_feats[s].append(1.0)
                else:
                    static_feats[s].append(0.0)

        # Ratio of current epoch and max. epochs
        for sf in range(len(static_feats)):
            static_feats[sf].append(epoch / max_epoch)

        if self.logger is not None:
            self.logger.info(f"Static features took {time.time() - static_time}\n\n")

        return np.array(static_feats)

    def dynamic_feature_gen(self, suggestions, data, predicted_quals, sm,
                            cot, results, next_instance_set):
        """Generate dynamic features.

        :param suggestions: list, suggested configurations
        :param data: data object, contains historic data
        :param predicted_quals: nested list, predicted performance/quality for
            suggested configurations
        :param sm: object, surrogates.SurrogateManager()
        :param cot: int, cut off time
        :param results: list, results of points evaluated so far
        :return dyn_feats: list, dynamic features
        """
        if self.logger is not None:
            all_dyn_time = time.time()
            dyn_one = time.time()

        # Features based on surrogates
        dyn_feats = self.expected_qual(suggestions, sm,
                                       cot, Surrogates.SMAC, next_instance_set)

        if self.logger is not None:
            self.logger.info(f"Dyn 1 features took {time.time() - dyn_one}\n\n")
            dyn_two = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                            self.expected_qual(suggestions, sm,
                                               cot, Surrogates.GGApp,
                                               None)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 2 features took {time.time() - dyn_two}\n\n")
            dyn_three = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                            self.expected_qual(suggestions, sm,
                                               cot, Surrogates.CPPL,
                                               next_instance_set)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 3 features took {time.time() - dyn_three}\n\n")
            dyn_four = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                            self.prob_qual_improve(suggestions, sm, cot,
                                                   results,
                                                   Surrogates.SMAC,
                                                   None)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 4 features took {time.time() - dyn_four}\n\n")
            dyn_five = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                            self.prob_qual_improve(suggestions, sm, cot,
                                                   results,
                                                   Surrogates.GGApp,
                                                   None)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 5 features took {time.time() - dyn_five}\n\n")
            dyn_six = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                            self.prob_qual_improve(suggestions, sm, cot,
                                                   results,
                                                   Surrogates.CPPL,
                                                   next_instance_set)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 6 features took {time.time() - dyn_six}\n\n")
            dyn_seven = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                            self.uncertainty_improve(suggestions, sm, cot,
                                                     Surrogates.SMAC,
                                                     None)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 7 features took {time.time() - dyn_seven}\n\n")
            dyn_eight = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                            self.uncertainty_improve(suggestions, sm, cot,
                                                     Surrogates.GGApp,
                                                     None)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 8 features took {time.time() - dyn_eight}\n\n")
            dyn_nine = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                            self.uncertainty_improve(suggestions, sm, cot,
                                                     Surrogates.CPPL,
                                                     next_instance_set)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 9 features took {time.time() - dyn_nine}\n\n")
            dyn_ten = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                           self.expected_improve(suggestions, sm, cot,
                                                 Surrogates.SMAC,
                                                 next_instance_set)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 10 features took {time.time() - dyn_ten}\n\n")
            dyn_eleven = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                           self.expected_improve(suggestions, sm, cot,
                                                 Surrogates.GGApp,
                                                 None)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Dyn 11 features took {time.time() - dyn_eleven}\n\n")
            dyn_twelve = time.time()

        dyn_feats = \
            np.concatenate((dyn_feats,
                           self.expected_improve(suggestions, sm, cot,
                                                 Surrogates.CPPL,
                                                 next_instance_set)),
                           axis=1)

        dyn_feats = self.surr_votes(dyn_feats)

        if self.logger is not None:
            self.logger.info(f"Dyn 12 features took {time.time() - dyn_twelve}\n\n")
            self.logger.info(f"All dyn features took {time.time() - all_dyn_time}\n\n")

        return np.array(dyn_feats)

    def diversity_feature_gen(self, suggestions, data, results, cot,
                              psetting, predicted_quals, evaluated):
        """Generate diversity features.

        :param suggestions: list, suggested configurations
        :param data: data object, contains historic data
        :param results: dict, qualities of configurations
        :param cot: float, cut off time for tournaments
        :param psetting: scenario.parameter
        :param predicted_quals: list, predicted qualities of
            points evaluated so far
        :param evaluated: list, all evaluated points so far
        :param sm: initialized surrogates.SurrogateManager()
        :return div_feats: list, diversity features
        """
        if self.logger is not None:
            div_time = time.time()

        nce = 0
        for content in data.values():
            nce += len(content.configuration_ids)

        generators = [gen for gen in Generator]

        # Features based on relatives evaluated so far
        div_feats = self.percent_rel_evals(suggestions, data, nce)
        div_feats = \
            np.concatenate((div_feats,
                            self.avg_rel_evals_qual(suggestions, data,
                                                    nce, results, cot,
                                                    generators)),
                           axis=1)
        div_feats = \
            np.concatenate((div_feats,
                            self.best_rel_evals_qual(suggestions, data,
                                                     generators, results,
                                                     cot)),
                           axis=1)
        div_feats = \
            np.concatenate((div_feats,
                            self.std_rel_evals_qual(suggestions, data,
                                                    generators, results,
                                                    cot)),
                           axis=1)

        div_feats = \
            np.concatenate((div_feats,
                           self.diff_pred_real_qual(suggestions, data,
                                                    predicted_quals,
                                                    results)),
                           axis=1)

        # Features based on points evaluated so far
        div_feats = \
            np.concatenate((div_feats,
                           self.avg_dist_evals(suggestions, evaluated,
                                               psetting)),
                           axis=1)
        div_feats = \
            np.concatenate((div_feats,
                           self.avg_dist_sel(suggestions, psetting)),
                           axis=1)
        div_feats = \
            np.concatenate((div_feats,
                           self.avg_dist_rel(suggestions, evaluated,
                                             psetting, generators)),
                           axis=1)

        if self.logger is not None:
            self.logger.info(f"Div features took {time.time() - div_time}\n\n")

        return np.array(div_feats)



"""This module contains functions for the GGA++ surrogate."""
import joblib
import six
import sklearn
import sklearn.ensemble._base as base
import sys
sys.modules['sklearn.ensemble.base'] = base
sys.modules['sklearn.externals.joblib'] = joblib
sys.modules['sklearn.externals.six'] = six
sys.modules['sklearn.externals.six.moves'] = six.moves

from costcla import CostSensitiveRandomForestClassifier  # noqa: E402
from costcla.models.cost_tree import CostSensitiveDecisionTreeClassifier  # noqa: E402
from costcla.models.bagging import BaggingClassifier  # noqa: E402
from costcla.metrics import cost_loss  # noqa: E402

import numpy as np  # noqa: E402
import numbers  # noqa: E402
import copy  # noqa: E402
import itertools  # noqa: E402
import statistics as st  # noqa: E402
from scipy.stats import norm  # noqa: E402
from sklearn.preprocessing import StandardScaler  # noqa: E402

from sklearn.ensemble import BaggingRegressor  # noqa: E402
from sklearn.ensemble._bagging import (  # noqa: E402
    BaseBagging,
    _parallel_build_estimators,
    _parallel_predict_regression
)
from sklearn.utils import check_random_state  # noqa: E402
from sklearn.utils.validation import (  # noqa: E402
    has_fit_parameter,
    check_is_fitted,
    _check_sample_weight
)
from sklearn.ensemble._base import BaseEnsemble, _partition_estimators  # noqa: E402
from sklearn.base import RegressorMixin  # noqa: E402
from sklearn.utils.fixes import delayed  # noqa: E402
from joblib import Parallel  # noqa: E402
MAX_INT = np.iinfo(np.int32).max

from selector.pool import ParamType, Generator  # noqa: E402
from selector.pool import Configuration as SelConfig  # noqa: E402
from selector.point_gen import PointGen  # noqa: E402
from selector.generators.default_point_generator import (  # noqa: E402
    check_conditionals,
    check_no_goods
)
from selector.generators.random_point_generator import (  # noqa: E402
    reset_no_goods,
    reset_conditionals,
    random_point
)
from selector.generators.variable_graph_point_generator import (  # noqa: E402
    variable_graph_point,
    Mode
)
from selector.point_gen import PointGen  # noqa: E402
import uuid  # noqa: E402
import random  # noqa: E402

from threadpoolctl import ThreadpoolController
controller = ThreadpoolController()


class GGApp(CostSensitiveDecisionTreeClassifier):
    """GGA++ Decision Tree Regressor."""

    def __init__(self,
                 criterion='direct_cost',
                 criterion_weight=False,
                 num_pct=10,
                 max_features=None,
                 max_depth=5,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_gain=.1,
                 pruned=True,
                 q=0.1
                 ):
        """Initialize Decision Tree Regressor."""
        self.criterion = criterion
        self.criterion_weight = criterion_weight
        self.num_pct = num_pct
        self.max_features = max_features
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_score = min_gain
        self.min_gain = min_gain
        self.pruned = pruned

        self.n_features_ = None
        self.max_features_ = None

        self.tree_ = []

        self.q = q

    def _node_value(self, y_true, X):
        """Private function to calculate the value of a node.

        Parameters
        ----------
        y_true : array indicator matrix
                Ground truth (correct) labels.
        Returns
        -------
        tuple(node prediction : float, node predicted probability : float)
        """
        # Criterion
        y_pred = st.median(y_true)
        y_prob = norm.pdf(3, loc=y_pred, scale=np.std(y_true))

        return y_pred, y_prob

    def fit(self, X, y, check_input=False):
        """Fitting function."""
        n_samples, self.n_features_ = X.shape

        # mock cost_matrix with no influence
        cost_mat = np.array([[1, 1, 0, 0] for _ in X])

        self.tree_ = self._tree_class()

        # Maximum number of features to be taken into account per split
        if isinstance(self.max_features, six.string_types):
            if self.max_features == "auto":
                max_features = max(1, int(np.sqrt(self.n_features_)))
            elif self.max_features == "sqrt":
                max_features = max(1, int(np.sqrt(self.n_features_)))
            elif self.max_features == "log2":
                max_features = max(1, int(np.log2(self.n_features_)))
            else:
                raise ValueError(
                    'Invalid value for max_features. Allowed string '
                    'values are "auto", "sqrt" or "log2".')
        elif self.max_features is None:
            max_features = self.n_features_
        elif isinstance(self.max_features, (numbers.Integral, np.integer)):
            max_features = self.max_features
        else:  # float
            if self.max_features > 0.0:
                max_features = max(
                    1, int(self.max_features * self.n_features_))
            else:
                max_features = 1  # On sklearn is 0.
        self.max_features_ = max_features

        self.tree_.tree = self._tree_grow(y, X, cost_mat)

        if self.pruned:
            self.pruning(X, y, cost_mat)

        return self

    def _tree_grow(self, y_true, X, cost_mat, level=0):

        if len(X.shape) == 1:
            tree = dict(y_pred=y_true, y_prob=0.5, level=level,
                        split=-1, n_samples=1, gain=0)
            return tree

        # Calculate the best split of the current node
        split, score, Xl_pred, y_pred, y_prob = self._best_split(
            y_true, X, cost_mat)

        n_samples, n_features = X.shape

        # Construct the tree object as a dictionary

        tree = dict(y_pred=y_pred, y_prob=y_prob, level=level,
                    split=-1, n_samples=n_samples, gain=score)

        # Check the stopping criteria
        if score < self.min_score:
            return tree
        if self.max_depth is not None:
            if level >= self.max_depth:
                return tree
        if n_samples <= self.min_samples_split:
            return tree
        if X.shape[0] <= 10:
            return tree

        j, l = split
        filter_Xl = (X[:, j] <= l)
        filter_Xr = ~filter_Xl
        n_samples_Xl = np.nonzero(filter_Xl)[0].shape[0]
        n_samples_Xr = np.nonzero(filter_Xr)[0].shape[0]

        if min(n_samples_Xl, n_samples_Xr) <= self.min_samples_leaf:
            return tree

        # No stopping criteria is met
        tree['split'] = split
        tree['node'] = self.tree_.n_nodes
        self.tree_.n_nodes += 1

        tree['sl'] = self._tree_grow(
            y_true[filter_Xl], X[filter_Xl], cost_mat[filter_Xl], level + 1)
        tree['sr'] = self._tree_grow(
            y_true[filter_Xr], X[filter_Xr], cost_mat[filter_Xr], level + 1)

        return tree

    def _best_split(self, y_true, X, cost_mat):

        n_samples, n_features = X.shape
        num_pct = self.num_pct

        cost_base, _, _ = self._node_cost(y_true, cost_mat)
        y_pred, y_prob = self._node_value(y_true, X)

        # Calculate the gain of all features each split in num_pct
        scores = np.zeros((n_features, num_pct))
        pred = np.zeros((n_features, num_pct))
        splits = np.zeros((n_features, num_pct))

        # Selected features
        selected_features = np.arange(0, self.n_features_)
        # Add random state
        np.random.shuffle(selected_features)
        selected_features = selected_features[:self.max_features_]
        selected_features.sort()

        # For each feature test all possible splits
        for j in selected_features:
            splits[j, :] = np.percentile(
                X[:, j], np.arange(0, 100, 100.0 / num_pct).tolist())

            for l in range(num_pct):
                # Avoid repeated values,
                # since np.percentile may return repeated values
                if l == 0 or (l > 0 and splits[j, l] != splits[j, l - 1]):
                    split = (j, splits[j, l])
                    scores[j, l], pred[j, l] = self._calculate_score(
                        cost_base, y_true, X, cost_mat, split, self.q)

        best_split = np.unravel_index(scores.argmax(), scores.shape)

        return (best_split[0], splits[best_split]), \
            scores.max(), pred[best_split], y_pred, y_prob

    def _calculate_score(self, cost_base, y_true, X, cost_mat, split, q):
        """Calculate GGA++ score."""
        # Get qth percentile performance threshold
        j, l = split
        X_sort = copy.deepcopy(X)
        y_sort = copy.deepcopy(y_true)
        sort = np.argsort(y_sort)
        # sort = np.argsort(y_sort)[::-1]
        y_sort = y_sort[sort]
        X_sort = X_sort[sort]
        h_idx = int(len(y_sort) * q)
        v_h = y_sort[h_idx]

        # Split by split value
        filter_Xl = (X_sort[:, j] <= l)  # L
        filter_Xr = ~filter_Xl  # R

        # T: all <= h_idx, U: all > h_idx

        # Compute score
        lls = 0
        rls = 0
        lts = 0
        rts = 0
        ltn = 0  # |L ∩ T|
        rtn = 0  # |R ∩ T|

        for l_idx, part_l in enumerate(filter_Xl):
            if l_idx > h_idx and part_l:
                lls += (v_h - y_true[l_idx])**2

            if l_idx <= h_idx and part_l:
                lts += (y_true[l_idx] - v_h)**2
                # ltn = |L ∩ T|
                ltn += 1

            if l_idx > h_idx and not part_l:
                rls += (v_h - y_true[l_idx])**2

            if l_idx <= h_idx and not part_l:
                rts += (y_true[l_idx] - v_h)**2
                # rtn = |R ∩ T|
                rtn += 1

        al = (ltn + lts) / (1 + lls)
        ar = (rtn + rts) / (1 + rls)

        if ltn > rtn:
            score = al
        elif ltn == rtn:
            score = min([al, ar])
        elif ltn < rtn:
            score = ar

        # Reduce Iterations avoiding low gain
        if score < self.min_gain:
            return 0.0, int(np.sign(y_true.mean() - 0.5) == 1)

        n_samples, n_features = X_sort.shape

        # Check if one of the leafs is empty
        # TODO: This must be check in _best_split
        if np.nonzero(filter_Xl)[0].shape[0] in [0, n_samples]:  # 1 leaf empty
            return 0.0, 0.0

        Xl_pred, _ = self._node_value(y_true[filter_Xr],
                                      cost_mat[filter_Xr, :])

        return score, Xl_pred


class GGAppRegressorMixin(RegressorMixin):
    """Mixin class for all regression estimators in scikit-learn."""

    _estimator_type = "regressor"

    def score(self, X, y, sample_weight=None):
        """Compute score."""
        from .metrics import r2_score

        y_pred = self.predict(X)
        return r2_score(y, y_pred, sample_weight=sample_weight)

    def _more_tags(self):
        return {"requires_y": True}

    def _fit(
        self,
        X,
        y,
        max_samples=None,
        max_depth=None,
        sample_weight=None,
        check_input=True,
    ):
        random_state = check_random_state(self.random_state)

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=None)

        # Remap output
        n_samples = X.shape[0]
        self._n_samples = n_samples
        y = self._validate_y(y)

        # Check parameters
        self._validate_estimator()

        if max_depth is not None:
            self.estimator_.max_depth = max_depth

        # Validate max_samples
        if max_samples is None:
            max_samples = self.max_samples
        elif not isinstance(max_samples, numbers.Integral):
            max_samples = int(max_samples * X.shape[0])

        if max_samples > X.shape[0]:
            raise ValueError("max_samples must be <= n_samples")

        # Store validated integer row sampling value
        self._max_samples = max_samples

        # Validate max_features
        if isinstance(self.max_features, numbers.Integral):
            max_features = self.max_features
        elif isinstance(self.max_features, float):
            max_features = int(self.max_features * self.n_features_in_)

        if max_features > self.n_features_in_:
            raise ValueError("max_features must be <= n_features")

        max_features = max(1, int(max_features))

        # Store validated integer feature sampling value
        self._max_features = max_features

        # Other checks
        if not self.bootstrap and self.oob_score:
            raise ValueError("Out of bag estimation only",
                             "available if bootstrap=True")

        if self.warm_start and self.oob_score:
            raise ValueError("Out of bag estimate only",
                             "available if warm_start=False")

        if hasattr(self, "oob_score_") and self.warm_start:
            del self.oob_score_

        if not self.warm_start or not hasattr(self, "estimators_"):
            # Free allocated memory, if any
            self.estimators_ = []
            self.estimators_features_ = []

        n_more_estimators = self.n_estimators - len(self.estimators_)

        if n_more_estimators < 0:
            raise ValueError(
                "n_estimators=%d must be larger or equal to "
                "len(estimators_)=%d when warm_start==True"
                % (self.n_estimators, len(self.estimators_))
            )

        elif n_more_estimators == 0:
            warn(
                "Warm-start fitting without increasing n_estimators does not "
                "fit new trees."
            )
            return self

        # Parallel loop
        n_jobs, n_estimators, starts = _partition_estimators(
            n_more_estimators, self.n_jobs
        )
        total_n_estimators = sum(n_estimators)

        # Advance random state to state after training
        # the first n_estimators
        if self.warm_start and len(self.estimators_) > 0:
            random_state.randint(MAX_INT, size=len(self.estimators_))

        seeds = random_state.randint(MAX_INT, size=n_more_estimators)
        self._seeds = seeds

        all_results = Parallel(
            n_jobs=n_jobs, verbose=self.verbose, **self._parallel_args()
        )(
            delayed(_parallel_build_estimators)(
                n_estimators[i],
                self,
                X,
                y,
                sample_weight,
                seeds[starts[i]: starts[i + 1]],
                total_n_estimators,
                verbose=self.verbose,
                check_input=check_input,
            )
            for i in range(n_jobs)
        )

        # Reduce
        self.estimators_ += list(
            itertools.chain.from_iterable(t[0] for t in all_results)
        )
        self.estimators_features_ += list(
            itertools.chain.from_iterable(t[1] for t in all_results)
        )

        if self.oob_score:
            self._set_oob_score(X, y)

        return self


class BaggingRegressor(GGAppRegressorMixin, BaseBagging):
    """sklearn Bagging Regressor Redefinition."""

    def __init__(
        self,
        n_estimators=10,
        *,
        max_samples=1.0,
        max_features=1.0,
        bootstrap=True,
        bootstrap_features=False,
        oob_score=False,
        warm_start=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        base_estimator="deprecated",
    ):
        """Initialize BaggingRegressor."""
        super().__init__(
            n_estimators=n_estimators,
            max_samples=max_samples,
            max_features=max_features,
            bootstrap=bootstrap,
            bootstrap_features=bootstrap_features,
            oob_score=oob_score,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            base_estimator=base_estimator,
        )

    def _fit(
        self,
        X,
        y,
        max_samples=None,
        max_depth=None,
        sample_weight=None,
        check_input=True,
    ):
        """Build a Bagging ensemble of estimators from the training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.
        y : array-like of shape (n_samples,)
            The target values (class labels in classification, real numbers in
            regression).
        max_samples : int or float, default=None
            Argument to use instead of self.max_samples.
        max_depth : int, default=None
            Override value used when constructing base estimator. Only
            supported if the base estimator has a max_depth parameter.
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted.
            Note that this is supported only if the base estimator supports
            sample weighting.
        check_input : bool, default=True
            Override value used when fitting base estimator. Only supported
            if the base estimator has a check_input parameter for fit function.
        Returns
        -------
        self : object
            Fitted estimator.
        """
        random_state = check_random_state(self.random_state)
        sample_weight = None

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=None)

        # Remap output
        n_samples = X.shape[0]
        self._n_samples = n_samples
        y = self._validate_y(y)

        # Check parameters
        self._validate_estimator()

        if max_depth is not None:
            self.estimator_.max_depth = max_depth

        # Validate max_samples
        if max_samples is None:
            max_samples = self.max_samples
        elif not isinstance(max_samples, numbers.Integral):
            max_samples = int(max_samples * X.shape[0])

        if max_samples > X.shape[0]:
            raise ValueError("max_samples must be <= n_samples")

        # Store validated integer row sampling value
        self._max_samples = max_samples

        # Validate max_features
        if isinstance(self.max_features, numbers.Integral):
            max_features = self.max_features
        elif isinstance(self.max_features, float):
            max_features = int(self.max_features * self.n_features_in_)

        if max_features > self.n_features_in_:
            raise ValueError("max_features must be <= n_features")

        max_features = max(1, int(max_features))

        # Store validated integer feature sampling value
        self._max_features = max_features

        # Other checks
        if not self.bootstrap and self.oob_score:
            raise ValueError("Out of bag estimation only",
                             "available if bootstrap=True")

        if self.warm_start and self.oob_score:
            raise ValueError("Out of bag estimate only",
                             " available if warm_start=False")

        if hasattr(self, "oob_score_") and self.warm_start:
            del self.oob_score_

        if not self.warm_start or not hasattr(self, "estimators_"):
            # Free allocated memory, if any
            self.estimators_ = []
            self.estimators_features_ = []

        n_more_estimators = self.n_estimators - len(self.estimators_)

        if n_more_estimators < 0:
            raise ValueError(
                "n_estimators=%d must be larger or equal to "
                "len(estimators_)=%d when warm_start==True"
                % (self.n_estimators, len(self.estimators_))
            )

        elif n_more_estimators == 0:
            warn(
                "Warm-start fitting without increasing n_estimators does not "
                "fit new trees."
            )
            return self

        # Parallel loop
        n_jobs, n_estimators, starts = _partition_estimators(
            n_more_estimators, self.n_jobs
        )
        total_n_estimators = sum(n_estimators)

        # Advance random state to state after training
        # the first n_estimators
        if self.warm_start and len(self.estimators_) > 0:
            random_state.randint(MAX_INT, size=len(self.estimators_))

        seeds = random_state.randint(MAX_INT, size=n_more_estimators)
        self._seeds = seeds

        all_results = Parallel(
            n_jobs=n_jobs, verbose=self.verbose, **self._parallel_args()
        )(
            delayed(_parallel_build_estimators)(
                n_estimators[i],
                self,
                X,
                y,
                sample_weight,
                seeds[starts[i]: starts[i + 1]],
                total_n_estimators,
                verbose=self.verbose,
                # check_input=check_input,
            )
            for i in range(n_jobs)
        )

        # Reduce
        self.estimators_ += list(
            itertools.chain.from_iterable(t[0] for t in all_results)
        )
        self.estimators_features_ += list(
            itertools.chain.from_iterable(t[1] for t in all_results)
        )

        if self.oob_score:
            self._set_oob_score(X, y)

        return self


class GGAppBaggingRegressor(BaggingRegressor):
    """GGA++ Bagging Regressor."""

    def __init__(
        self,
        n_estimators=10,
        *,
        max_samples=1.0,
        max_features=1.0,
        bootstrap=True,
        bootstrap_features=False,
        oob_score=False,
        warm_start=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        base_estimator="deprecated",
    ):
        """Initialize GGA++ Bagging Regressor."""
        super().__init__(
            n_estimators=n_estimators,
            max_samples=max_samples,
            max_features=max_features,
            bootstrap=bootstrap,
            bootstrap_features=bootstrap_features,
            oob_score=oob_score,
            warm_start=warm_start,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            base_estimator=base_estimator,
        )

    def _set_oob_score(self, X, y):
        n_samples = y.shape[0]

        predictions = np.zeros((n_samples,))
        n_predictions = np.zeros((n_samples,))

        for estimator, samples, features in zip(
            self.estimators_, self.estimators_samples_,
            self.estimators_features_
        ):
            # Create mask for OOB samples
            mask = ~indices_to_mask(samples, n_samples)

            predictions[mask] += \
                estimator.predict((X[mask, :])[:, features])
            n_predictions[mask] += 1

        if (n_predictions == 0).any():
            warn(
                "Some inputs do not have OOB scores. "
                "This probably means too few estimators were used "
                "to compute any reliable oob estimates."
            )
            n_predictions[n_predictions == 0] = 1

        predictions /= n_predictions

        self.oob_prediction_ = predictions
        self.oob_score_ = r2_score(y, predictions)

    def predict(self, X):
        """Predict regression target for X.

        The predicted regression target of an input sample is computed as the
        mean predicted regression targets of the estimators in the ensemble.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Sparse matrices are accepted only if
            they are supported by the base estimator.
        Returns
        -------
        y : ndarray of shape (n_samples,)
            The predicted values.
        """
        check_is_fitted(self)
        # Check data
        X = self._validate_data(
            X,
            accept_sparse=["csr", "csc"],
            dtype=None,
            force_all_finite=False,
            reset=False,
        )

        # Parallel loop
        n_jobs, _, starts = _partition_estimators(self.n_estimators,
                                                  self.n_jobs)

        all_y_hat = Parallel(n_jobs=n_jobs, verbose=self.verbose)(
            delayed(_parallel_predict_regression)(
                self.estimators_[starts[i]: starts[i + 1]],
                self.estimators_features_[starts[i]: starts[i + 1]],
                X,
            )
            for i in range(n_jobs)
        )

        # Reduce
        y_hat = sum(all_y_hat) / self.n_estimators

        return y_hat


class GGAppRandomForestRegressor(GGAppBaggingRegressor):
    """GGA++ Random Forest."""

    def __init__(self,
                 n_estimators=10,
                 combination='majority_voting',
                 max_features='auto',
                 n_jobs=1, # will be scenario.tournament_size
                 verbose=False,
                 pruned=False,
                 q=0.1):
        """Initialize Random Forest."""
        super(GGAppBaggingRegressor, self).__init__(
            base_estimator=GGApp(max_features=max_features,
                                 pruned=pruned, q=q),
            n_estimators=n_estimators,
            max_samples=1.0,
            max_features=1.0,
            bootstrap=True,
            bootstrap_features=False,
            n_jobs=n_jobs,
            random_state=None,
            verbose=verbose)
        self.pruned = pruned


class GGAppSurr():
    """Surrogate GGA++."""

    def __init__(self, scenario, seed=False, cost=[1, 1, 0, 0], logger=None):
        """Initialize GGA++ surrogate.

        :param scenario: dict, scenario translated for GGA++
        :param seed: int, random seed
        """
        if not seed:
            self.seed = False
        else:
            self.seed = seed

        # Control How many threads/cores numpy and scipy use
        @controller.wrap(limits=scenario.tournament_size,
                         user_api='openmp')
        @controller.wrap(limits=scenario.tournament_size,
                         user_api='blas')
        def threaded_init(scenario, seed, cost):
            self.scenario = scenario
            self.logger = logger
            self.regressor = \
                GGAppRandomForestRegressor(n_jobs=scenario.tournament_size)
            self.sc = StandardScaler()
            self.x_stash = np.array([])
            self.y_stash = np.array([])
            self.cost = cost
            self.best_q = None

            self.transfom_selector_scenario_for_ggapp(scenario)
            self.random_generator = PointGen(self.scenario, random_point,
                                             seed=self.seed)
            self.variable_graph_generator = PointGen(self.scenario,
                                                     variable_graph_point,
                                                     seed=42)

        threaded_init(scenario, seed, cost)

    def transfom_selector_scenario_for_ggapp(self, scenario):
        """Save parameter types.

        :param scenario: scenario object from selector
        """
        paramsecenario = self.scenario.parameter
        self.types = {}
        for ps in paramsecenario:
            self.types[ps.name] = ps.type

    def set_cat(self, c):
        """Set Cat value to binary."""
        if c is True:
            c = 1
        elif c is False:
            c = 0

        return c

    def transform_values(self, conf):
        """Transform configuration values in GGA++ format.

        :param conf: object, selector.pool.Configuration
        :return config: dict, parameter values
        """
        config = []
        count = 0
        if type(conf) is list:
            for c in conf:
                for param in self.scenario.parameter:
                    if param.name not in c.conf:
                        c.conf[param.name] = param.bound[0]
            for c in conf:
                count += 1
                single_conf = []
                for t in self.types.keys():
                    if self.types[t] == ParamType.categorical and t in c.conf:
                        if c.conf[t] is None:
                            for sp in self.scenario.parameter:
                                if t == sp.name:
                                    single_conf.append(sp.bound[0])
                        else:
                            if isinstance(c.conf[t], (str, np.str_)):
                                for sp in self.scenario.parameter:
                                    if t == sp.name:
                                        single_conf.append(
                                            sp.bound.index(c.conf[t]))
                            else:
                                single_conf.append(
                                    self.set_cat(float(c.conf[t])))
                    elif t in c.conf:
                        if c.conf[t] is None:
                            for sp in self.scenario.parameter:
                                if t == sp.name:
                                    single_conf.append(sp.bound[0])
                        else:
                            single_conf.append(float(c.conf[t]))
                config.append(np.array(single_conf))
        else:
            for param in self.scenario.parameter:
                if param.name not in conf.conf:
                    conf.conf[param.name] = param.bound[0]
            for t in self.types.keys():
                if self.types[t] == ParamType.categorical and t in conf.conf:
                    if conf.conf[t] is None:
                        for sp in self.scenario.parameter:
                            if t == sp.name:
                                config.append(sp.bound[0])
                    else:
                        if isinstance(conf.conf[t], (str, np.str_)):
                            for sp in self.scenario.parameter:
                                if t == sp.name:
                                    config.append(
                                        sp.bound.index(conf.conf[t]))
                        else:
                            config.append(self.set_cat(float(conf.conf[t])))
                elif t in conf.conf:
                    if conf.conf[t] is None:
                        for sp in self.scenario.parameter:
                            if t == sp.name:
                                config.append(sp.bound[0])
                    else:
                        config.append(float(conf.conf[t]))

        config = np.array(config, dtype=object)

        return config

    def get_costs(self, y):
        """Generate cost matrix."""
        costs = np.array([self.cost for _ in y])

        return costs

    def update(self, history, configs, results, terminations, ac_runtime=None):
        """Update GGA++ epm.

        :param results: results of the Tournament
        :param conf: object, selector.pool.Configuration
        """
        conf = []
        result = []
        config_dict = {}
        for c in configs:
            config_dict[c.id] = c

        # instances in tournament
        instances = history.instance_set
        for cid in config_dict.keys():
            use_results = False
            perf_sum = 0
            perf_count = 0
            # config in results
            for ins in instances:
                # OMIT every censored date in update
                if cid in terminations:
                    if ins in terminations[cid]:
                        break
                else:
                    r = results[cid][ins]
                    perf_count += 1
                    use_results = True
                    if r is not None and not np.isnan(r):
                        perf_sum += results[cid][ins]
                        # result.append(results[cid][ins])
                    else:
                        # This cid/ins pair was a time limit reach
                        perf_sum += self.scenario.cutoff_time
                        # result.append(self.scenario.cutoff_time)
            if use_results:
                result.append(perf_sum / perf_count)
                conf.append(self.transform_values(config_dict[cid]))

        if ac_runtime >= self.scenario.wallclock_limit * 0.15 and \
                len(self.y_stash) > len(result) * 2:
            self.y_stash = self.y_stash[len(result):]
            self.y_stash = np.append(self.y_stash, np.array(result))
        else:
            self.y_stash = np.append(self.y_stash, np.array(result))

        if self.best_q is None:
            if len(self.y_stash) == 0:
                import sys
                self.best_q = sys.maxsize
            else:
                self.best_q = np.min(self.y_stash)
        elif np.min(self.y_stash) < self.best_q:
            self.best_q = np.min(self.y_stash)

        if len(conf) != 0:

            if len(self.x_stash) > 0:
                if ac_runtime >= self.scenario.wallclock_limit * 0.15 and \
                        len(self.x_stash) > len(conf) * 2:
                    self.x_stash = self.x_stash[len(conf):]
                    self.x_stash = np.vstack([self.x_stash, np.array(conf)])
                else:
                    self.x_stash = np.vstack([self.x_stash, np.array(conf)])
            else:
                self.x_stash = np.array(conf)

            self.x_stash = self.sc.fit_transform(self.x_stash)
            self.regressor.fit(self.x_stash, self.y_stash,
                               self.get_costs(self.y_stash))

        if self.logger is not None:
            self.logger.info(f"Length x_stash: {len(self.x_stash)}")

    def get_suggestions(self, scenario, n_samples, data, results, _,
                        oversampling=10):
        """Get point suggestions from GGA++.

        :param scenario: scenario object from selector
        :param n_samples: int, number of point suggestions to generate
        :return suggestions: list, list of configurations
        """
        suggestions = []
        for i in range(oversampling * n_samples):
            suggestions.append(self.variable_graph_generator.point_generator(
                results=results, mode=Mode.best_and_random,
                alldata=data, lookback=i + 1, seed=(42 + i)))

        if len(self.x_stash) > 1:
            predicted_quality = \
                self.predict(suggestions)[0]
            sugg_sorted = np.argsort(predicted_quality)
        else:
            sugg_sorted = [i for i in range(n_samples)]

        best_idx = sugg_sorted[:n_samples]

        best_suggs = list(np.array(suggestions)[best_idx])[:n_samples]

        suggestions = []

        for idx, bs in enumerate(best_suggs):
            config_setting = {}

            if not self.seed:
                identity = uuid.uuid4()
            else:
                identity = uuid.UUID(int=random.getrandbits(self.seed))

            for t in self.types.keys():
                if t in bs.conf:
                    config_setting[t] = bs.conf[t]

            suggestions.append(SelConfig(identity,
                                         config_setting,
                                         Generator.ggapp))

            # Check conditionals and reset parameters if violated
            cond_vio = check_conditionals(scenario,
                                          suggestions[idx].conf)
            if cond_vio:
                suggestions[idx].conf = \
                    reset_conditionals(scenario,
                                       suggestions[idx].conf,
                                       cond_vio)

            # Check no goods and reset values if violated
            ng_vio = check_no_goods(scenario, suggestions[idx].conf)
            while ng_vio:
                suggestions[idx].conf = \
                    reset_no_goods(scenario, suggestions[idx].conf)
                ng_vio = check_no_goods(scenario,
                                        suggestions[idx].conf)

        return suggestions

    def predict(self, confs, _=None):
        """Predict performance/quality of configurations with GGA++ epm.

        :param confs: list of objects, [selector.pool.Configuration,]
        :return [mean, var]: numpy.ndarray, mean and variance of predicted
            performance/quality
        """
        if type(confs) is not np.ndarray:
            confs = self.transform_values(confs)

        confs = self.sc.transform(confs)
        estimators_predicts = \
            np.array(([x.predict(confs) for x in self.regressor.estimators_]))

        variances = []
        for idx, _ in enumerate(estimators_predicts[0]):
            variances.append(np.var(estimators_predicts[:, idx]))

        return np.array(self.regressor.predict(confs)), np.array(variances)

    def expected_improvement(self, suggestions, _):
        """Compute expected improvement via GGA++ model.

        :param suggestions: list of objects, [selector.pool.Configuration,]
        :return ei: numpy.ndarray, expected improvements
        """
        mean, var = self.predict(suggestions)
        std = np.sqrt(var)

        def calculate_ei():
            z = (self.best_q - mean) / std
            return (self.best_q - mean) * norm.cdf(z) + std * norm.pdf(z)

        if np.any(std == 0.0):
            stdc = np.copy(std)
            std[stdc == 0.0] = 1.0
            ei = calculate_ei()
            ei[stdc == 0.0] = 0.0
            return ei
        else:
            return calculate_ei()

    def probability_improvement(self, suggestions, r, ni):
        """Compute probability of improvement.

        :param suggestions: list of objects, [selector.pool.Configuration,]
        :return pi: numpy.ndarray, probabilities of improvement
        """
        mean, var = self.predict(suggestions)
        std = np.sqrt(var)

        pi = norm.cdf((self.best_q - mean) / std)
        pi_output = []
        for p in pi:
            if np.isnan(p):
                pi_output.append([0])
            else:
                pi_output.append([p])

        return pi_output



import os
import warnings
import argparse
import sys
sys.path.append(os.getcwd())

from selector.read_files import get_ta_arguments_from_pcs, read_instance_paths, read_instance_features


class Scenario:

    def __init__(self, scenario, cmd={'check_path': False}):
        """
        Scenario class that stores all relevant information for the configuration
        :param scenario: dic or string. If string, a scenario file will be read in.
        :param cmd: dic, Command line arguments which augment the scenario file/dic
        """

        if isinstance(scenario, str):
            scenario = self.scenario_from_file(scenario)

        elif isinstance(scenario, dict):
            scenario = scenario

        else:
            raise TypeError("Scenario must be string or dic")

        # add and overwrite cmd line args
        for key, value in cmd.items():

            if key in scenario and value is not None:
                warnings.warn(f"Setting: {key} of the scenario file is overwritten by parsed command line arguments")
                scenario[key] = value

            elif key not in scenario:
                scenario[key] = value

        self.read_scenario_files(scenario)

        for arg_name, arg_value in scenario.items():
            setattr(self, arg_name, arg_value)

        self.cutoff_time = float(self.cutoff_time)
        self.wallclock_limit = float(self.wallclock_limit)

        self.verify_scenario()

        # with open(f'./selector/logs/{self.log_folder}/scenario.pkl', 'wb') as out:
        #    pickle.dump(scenario, out)

    def read_scenario_files(self, scenario):

        """
        Read in the relevant files needed for a complete scenario
        :param scenario: dic.
        :return: scenario: dic.
        """

        # read in
        if "paramfile" in scenario:
            scenario["parameter"], scenario["no_goods"], scenario["conditionals"] = get_ta_arguments_from_pcs(scenario["paramfile"])
        else:
            raise ValueError("Please provide a file with the target algorithm parameters")

        if "instance_file" in scenario:
            scenario["instance_set"] = read_instance_paths(scenario["instance_file"])
        else:
            raise ValueError("Please provide a file with the training instances")

        if "test_instance_file" in scenario:
            scenario["test_instances"] = read_instance_paths(scenario["test_instance_file"])
        else:
            scenario["test_instances"] = []

        if "feature_file" in scenario:
            scenario["features"], scenario["feature_names"] = read_instance_features(scenario["feature_file"])
        else:
            raise ValueError("Please provide a file with instance features")

        return scenario

    def verify_scenario(self):
        """
        Verify that the scenario attributes are valid
        """
        # TODO: verify algo and execdir

        if self.run_obj not in ["runtime", "quality"]:
            raise ValueError("The specified run objective is not supported")

        if self.overall_obj not in ["mean", "mean10", "PAR10"]:
            raise ValueError("The specified objective is not supported")

        if not isinstance(float(self.cutoff_time), float):
            raise ValueError("The cutoff_time needs to be a float")

        if not isinstance(float(self.wallclock_limit), float):
            raise ValueError("The wallclock_limit needs to be a float")

        # check if the named instances are really available
        if self.check_path:
            for i in (self.instance_set + self.test_instances):
                if not os.path.exists(f"./selector{i}".strip("\n")):
                    raise FileExistsError(f"Instance file {i} does not exist")

        for i in (self.instance_set + self.test_instances):
            if i not in self.features:
                raise ValueError(f"For instance {i} no features were provided")

        if "log_folder" not in list(self.__dict__.keys()):
            setattr(self, "log_folder", "latest")
        elif self.log_folder == "None":
            self.log_folder = "latest"

    def scenario_from_file(self, scenario_path):

        """
        Read in an ACLib scenario file
        :param scenario_path: Path to the scenario file
        :return: dic containing the scenario information
        """

        name_map = {"algo": "ta_cmd"}
        scenario_dict = {}

        with open(scenario_path, 'r') as sc:
            for line in sc:
                line = line.strip()

                if "=" in line:

                    #remove comments
                    pairs = line.split("#", 1)[0].split("=")
                    pairs = [l.strip(" ") for l in pairs]

                    # change of AClib names to names we use. Extend name_map if necessary
                    if pairs[0] in name_map:
                        key = name_map[pairs[0]]
                    else:
                        key = pairs[0]

                    scenario_dict[key] = pairs[1]
        return scenario_dict


class LoadOptionsFromFile (argparse.Action):
    def __call__ (self, parser, namespace, values, option_string=None):
        with values as f:
            parser.parse_args(f.read().split(), namespace)


def parse_args():
    """
    Argparser
    :return: dic. Dic of arguments parsed
    """

    parser = argparse.ArgumentParser()
    hp = parser.add_argument_group("Hyperparameters of selector")
    so = parser.add_argument_group("Scenario options")

    hp.add_argument('--file', type=open, action=LoadOptionsFromFile)

    #parser.add_argument('--check_path', dest='check_path', action='store_true')
    parser.add_argument('--check_path', default=False, type=lambda x: (str(x).lower() == 'true'))

    hp.add_argument('--seed', default=42, type=int)
    hp.add_argument('--ta_pid_name', type=str, default="")
    hp.add_argument('--log_folder', type=str, default="latest")
    hp.add_argument('--memory_limit', type=int, default=1023*3)

    hp.add_argument('--ta_run_type', type=str, default="import_wrapper")
    hp.add_argument('--wrapper_mod_name', type=str, default="")
    hp.add_argument('--wrapper_class_name', type=str, default="")
    hp.add_argument('--quality_match', type=str, default="")
    hp.add_argument('--solve_match', nargs='+', type=str, default="")
    hp.add_argument('--quality_extract', type=str, default="")

    hp.add_argument('--winners_per_tournament', type=int, default=1)
    hp.add_argument('--tournament_size', type=int, default=5)
    hp.add_argument('--number_tournaments', type=int, default=2)

    hp.add_argument('--par', type=int, default=1)
    hp.add_argument('--monitor', type=str, default="tournament_level")
    hp.add_argument('--surrogate_amortized_time', type=int, default=30)

    hp.add_argument('--termination_criterion', type=str, default="runtime")
    hp.add_argument('--total_tournament_number', type=int, default=10)
    hp.add_argument('--model_update_iteration', type=int, default=3)

    hp.add_argument('--generator_multiple', type=int, default=5)
    hp.add_argument('--initial_instance_set_size', type=int, default=5)
    hp.add_argument('--set_size', type=int, default=50)
    hp.add_argument('--instances_dir', type=str, default="")
    hp.add_argument('--smac_pca_dim', type=int, default=8)

    so.add_argument('--scenario_file', type=str)
    so.add_argument('--ta_cmd', type=str)
    so.add_argument('--deterministic', type=str)
    so.add_argument('--run_obj', type=str)
    so.add_argument('--overall_obj', type=str)
    so.add_argument('--cutoff_time', type=str)
    so.add_argument('--wallclock_limit', type=str)
    so.add_argument('--instance_file', type=str)
    so.add_argument('--feature_file', type=str)
    so.add_argument('--paramfile', type=str)
    so.add_argument('--qual_max', type=bool, default=False)
    so.add_argument('--output_trigger', action='store_true')
    so.set_defaults(output_trigger=False)

    so.add_argument('--w_1', type=float, default=-0.8356679356095191)
    so.add_argument('--w_2', type=float, default=0.8572501015854599)
    so.add_argument('--w_3', type=float, default=0.6037123781208655)
    so.add_argument('--w_4', type=float, default=0.7357117939046054)
    so.add_argument('--w_5', type=float, default=-0.9936592470722203)
    so.add_argument('--w_6', type=float, default=-0.6555611254096128)
    so.add_argument('--w_7', type=float, default=0.6846488604171828)
    so.add_argument('--w_8', type=float, default=-0.29616131125825584)
    so.add_argument('--w_9', type=float, default=-0.42641293642609523)
    so.add_argument('--w_10', type=float, default=-0.636625827266299)
    so.add_argument('--w_11', type=float, default=-0.07705211904734346)
    so.add_argument('--w_12', type=float, default=0.7655061394605909)
    so.add_argument('--w_13', type=float, default=0.07958609777734993)
    so.add_argument('--w_14', type=float, default=-0.6433479035364913)
    so.add_argument('--w_15', type=float, default=0.9330160438280031)
    so.add_argument('--w_16', type=float, default=0.19269241833198825)
    so.add_argument('--w_17', type=float, default=0.924764738193058)
    so.add_argument('--w_18', type=float, default=0.6163122179339903)
    so.add_argument('--w_19', type=float, default=-0.3164211654229764)
    so.add_argument('--w_20', type=float, default=-0.9134468738753907)
    so.add_argument('--w_21', type=float, default=0.09507515338942374)
    so.add_argument('--w_22', type=float, default=-0.12320139765942552)
    so.add_argument('--w_23', type=float, default=0.5334762956128223)
    so.add_argument('--w_24', type=float, default=0.9040962774172033)
    so.add_argument('--w_25', type=float, default=0.7152291148005006)
    so.add_argument('--w_26', type=float, default=0.6264909195211199)
    so.add_argument('--w_27', type=float, default=-0.3078428587569706)
    so.add_argument('--w_28', type=float, default=0.13991257638679933)
    so.add_argument('--w_29', type=float, default=0.6072035854383455)
    so.add_argument('--w_30', type=float, default=0.07142940503717064)
    so.add_argument('--w_31', type=float, default=0.9481229658781036)
    so.add_argument('--w_32', type=float, default=0.260083514586567)
    so.add_argument('--w_33', type=float, default=-0.18358836077393523)
    so.add_argument('--w_34', type=float, default=0.5345236666392377)
    so.add_argument('--w_35', type=float, default=0.7620279025679817)
    so.add_argument('--w_36', type=float, default=-0.19720345812081364)
    so.add_argument('--w_37', type=float, default=0.34488933380271597)
    so.add_argument('--w_38', type=float, default=-0.31776509157739097)
    so.add_argument('--w_39', type=float, default=-0.060524835116740255)
    so.add_argument('--w_40', type=float, default=-0.04932352546335972)
    so.add_argument('--w_41', type=float, default=0.8845189709705542)
    so.add_argument('--w_42', type=float, default=0.07230604126060913)
    so.add_argument('--w_43', type=float, default=-0.6314381312161101)
    so.add_argument('--w_44', type=float, default=-0.2997826811933294)
    so.add_argument('--w_45', type=float, default=0.3920074130279137)
    so.add_argument('--w_46', type=float, default=0.7955255094472115)
    so.add_argument('--w_47', type=float, default=-0.9803992172559881)
    so.add_argument('--w_48', type=float, default=0.8009359266585256)

    return vars(parser.parse_args())


if __name__ == "__main__":

    parser = parse_args()



"""This module contains the latin hyper cube graph point generator."""

from skopt.space import Space
from skopt.sampler import Lhs
from enum import Enum
import uuid
from selector.pool import Configuration, ParamType, Generator
from selector.generators.default_point_generator import (
    check_conditionals,
    check_no_goods
)
from selector.generators.random_point_generator import reset_no_goods,\
    reset_conditionals


class LHSType(Enum):
    """Contains the types of LHS."""

    classic = 1
    centered = 2


class Criterion(Enum):
    """Contains the criterions of optimization of LHS."""

    correlation = 1
    maximin = 2
    ratio = 3


def generate_space(s):
    """
    Generating the sampling space for the lhc according to scenario parameters.

    : param scenario: scenario object
    """
    space_list = []
    for ps in s.parameter:
        if ps.type == ParamType.categorical:
            # categorical space defined by list
            space_list.append(ps.bound)
        else:
            # int/real space defined by tuple
            space_list.append(tuple(ps.bound))

    space = Space(space_list)

    return space


def get_n_points(space, n_samples, seed, lhs_type, criterion):
    """
    Generate n samples.

    : param space: sampling space for the lhc
    : param n_samples: number of samples to generate
    : param seed: will set random seed, if not False
    : param lhs_type: sampling type parameter for skopt.sampler.Lhs
    : param criterion: optimization criterion for skopt.sampler.Lhs
    return: n samples
    """
    if lhs_type == LHSType.centered:
        lt = 'centered'
    else:
        lt = 'classic'

    if criterion == Criterion.correlation:
        cr = 'correlation'
    elif criterion == Criterion.maximin:
        cr = 'maximin'
    elif criterion == Criterion.ratio:
        cr = 'ratio'
    else:
        cr = None

    lhs = Lhs(lhs_type=lt, criterion=cr)

    if seed:
        n_samples = lhs.generate(space.dimensions, n_samples,
                                 random_state=seed)
    else:
        n_samples = lhs.generate(space.dimensions, n_samples)

    return n_samples


def lhc_points(s, identity, n_samples=1, seed=False, lhs_type=LHSType.classic,
               criterion=None):
    """
    Configuration is generated via variable graph method.

    : param s: scenario object
    : param identity: uuid to identify configuration
    : param n_samples: int, number of picks from parameter space
    return: n configurations
    """
    space = generate_space(s)

    n_samples = get_n_points(space, n_samples, seed, lhs_type, criterion)

    param_names = []
    for param in s.parameter:
        param_names.append(param.name)

    n_points = []
    for sample in n_samples:
        point = {}
        for i in range(len(sample)):
            point[param_names[i]] = sample[i]
        n_points.append(point)

    # Check no goods and reset values if violated
    for point in n_points:
        ng_vio = check_no_goods(s, point)
        while ng_vio:
            point = reset_no_goods(s, point)
            ng_vio = check_no_goods(s, point)

    n_configurations = []

    if len(n_points) > 1:
        mult = True

    for conf in n_points:
        if mult:
            identity = uuid.uuid4()
        n_configurations.append(Configuration(identity, conf, Generator.lhc))

    return n_configurations



"""
This module sorts the configurations by performance within an iteration
and declares the overall best and saves it in a file.
"""

import json
import math


def read_run_hstory(path, timelimit):
    """
    This function reads in the run history.
    """
    
    # Opening JSON files
    f = open(f'{path}run_history.json')
    rh = json.load(f)
    f.close()

    with open(f'{path}trajectory.json', 'r') as handle:
        json_data = [json.loads(line) for line in handle]
    last_entry = list(json_data[-1].keys())[0]
    last_perf = 0
    for lp in list(rh[last_entry].values()):
        if math.isnan(lp):
            last_perf += timelimit
        else:
            last_perf += lp
    last_perf = last_perf / len(rh[last_entry])
    tr = {}
    for line in json_data:
        tr[list(line.keys())[0]] = list(line.values())[0]

    f.close()

    return rh, tr, last_entry, last_perf


def compute_performances(path, timelimit):
    """Compute and sort by performances."""

    performances = {}
    data, tr, last_entry, last_perf = read_run_hstory(path, timelimit)

    for k, v in data.items():
        len_v = len(v)
        if len_v not in performances:
            performances[len_v] = {}
        avg_perf = 0
        for perf in v.values():
            if math.isnan(perf):
                avg_perf += timelimit
            else:
                avg_perf += float(perf)
        performances[len_v][k] = avg_perf / len_v

    for perf in performances.keys():
        performances[perf] = \
            dict(sorted(performances[perf].items(), key=lambda x: x[1]))

    return dict(sorted(performances.items())), tr, data, last_entry, last_perf


def safe_best(path, timelimit):
    """Safe performances dict nd overall best config."""

    perfs, tr, data, best, last_perf = \
        compute_performances(path, timelimit)

    potential_best = list(perfs[list(perfs.keys())[-1]].keys())[0]

    while list(perfs.keys()) and potential_best != best:
        potential_best = list(perfs[list(perfs.keys())[-1]].keys())[0]
        if last_perf > perfs[list(perfs.keys())[-1]][potential_best]:
            if all(instance in data[best]
                    for instance in data[potential_best]) \
                    and potential_best in tr:
                potential_perf = []
                for instance in data[best]:
                    if math.isnan(data[potential_best][instance]):
                        potential_perf.append(timelimit)
                    else:
                        potential_perf.append(data[potential_best][instance])
                if sum(potential_perf) / len(potential_perf) < last_perf:
                    best = potential_best
                else:
                    del perfs[list(perfs.keys())[-1]][potential_best]
                    if len(perfs[list(perfs.keys())[-1]]) == 0:
                        del perfs[list(perfs.keys())[-1]]
            else:
                del perfs[list(perfs.keys())[-1]][potential_best]
                if len(perfs[list(perfs.keys())[-1]]) == 0:
                    del perfs[list(perfs.keys())[-1]]
        elif last_perf == perfs[list(perfs.keys())[-1]][potential_best]:
            break
        else:
            del perfs[list(perfs.keys())[-1]]

    overall_best = \
        {best: {'conf': tr[best], 'avg_perf': list(perfs.values())[-1][best]}}

    with open(f'{path}ranked_performances.json', 'w') as f:
        json.dump(perfs, f)

    with open(f'{path}overall_best.json', 'w') as f:
        json.dump(overall_best, f)

    return overall_best


if __name__ == "__main__":
    pass
